#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""This module contains the base taurus name validator classes"""
from __future__ import print_function

import re
from taurus import tauruscustomsettings
from taurus.core.util.singleton import Singleton
from taurus.core.taurushelper import makeSchemeExplicit

__all__ = ["TaurusAuthorityNameValidator", "TaurusDeviceNameValidator",
           "TaurusAttributeNameValidator"]


__docformat__ = "restructuredtext"


class _TaurusBaseValidator(Singleton):
    '''This is a private base class for taurus base validators. Do not derive
    from it if you are implementing a new scheme. Derive from the public
    classes from this module instead.
    '''
    scheme = None
    auth = '(?!)'  # note: '(?!)' is a pattern that will never match
    path = '(?!)'
    query = '(?!)'
    fragment = '(?!)'

    def __init__(self):
        if self.scheme is None:
            msg = ('This is  an abstract name validator class. ' +
                   'Only scheme-specific derived classes can be instantiated')
            raise NotImplementedError(msg)

        self.name_re = re.compile(self.namePattern)
        if self.nonStrictNamePattern is not None:
            self.nonStrictName_re = re.compile(self.nonStrictNamePattern)
        else:
            self.nonStrictName_re = None

    @property
    def namePattern(self):
        '''Provides a name pattern by composing the pattern strings for the
        URI segments'''
        return self.pattern % dict(scheme=self.scheme,
                                   authority=self.authority,
                                   path=self.path,
                                   query=self.query,
                                   fragment=self.fragment)

    @property
    def nonStrictNamePattern(self):
        '''implement in derived classes if a "less strict" pattern is allowed
        (e.g. for backwards-compatibility, "tango://a/b/c" could be an accepted
        device name, even if it breaks RFC3986).
        '''
        return None

    def isValid(self, name, matchLevel=None, strict=None):
        '''Whether the name matches the validator pattern.
        If strict is False, it also tries to match against the non-strict regexp
        (It logs a warning if it matched only the non-strict alternative)

        .. note:: The "matchLevel" keyword argument is deprecated and only
                  implemented for backwards compatibility. Do not use it for
                  new classes
        '''
        # warn if the deprecated matchLevel kwarg was received
        if matchLevel is not None:
            return self._isValidAtLevel(name, matchLevel=matchLevel)
        return self.getUriGroups(name, strict=strict) is not None

    def _isValidAtLevel(self, name, matchLevel=None):
        # matchLevel is a tango-centric deprecated  argument of isValid. Warn.
        msg = ('matchLevel is a Tango-centric concept. Avoid it outside ' +
               'the tango scheme')
        from taurus import warning
        warning(msg)
        return self.isValid(name)

    def getUriGroups(self, name, strict=None):
        '''returns the named groups dictionary from the URI regexp matching.
        If strict is False, it also tries to match against the non-strict regexp
        (It logs a warning if it matched only the non-strict alternative)
        '''
        if strict is None:
            strict = getattr(tauruscustomsettings, 'STRICT_MODEL_NAMES', False)
        name = makeSchemeExplicit(name, default=self.scheme)
        m = self.name_re.match(name)
        # if it is strictly valid, return the groups
        if m is not None:
            ret = m.groupdict()
            ret['__STRICT__'] = True
            return ret
        # if we are strict (or no less-strict pattern is defined) return None
        if strict or self.nonStrictName_re is None:
            return None
        # If a less-strict pattern is defined, use it, but warn if it works
        m = self.nonStrictName_re.match(name)
        if m is None:
            return None
        else:
            from taurus import warning
            msg = ('Model name "%s" is supported but not strictly valid. \n' +
                   'It is STRONGLY recommended that you change it to \n' +
                   'strictly follow %s scheme syntax') % (name, self.scheme)
            warning(msg)
            ret = m.groupdict()
            ret['__STRICT__'] = False
            return ret

    def getParams(self, name):
        # deprecation warning
        msg = ('%s.getParams() is deprecated. Use getUriGroups() instead.' %
               self.__class__.__name__)
        from taurus import warning
        warning(msg)
        # support old group names
        groups = self.getUriGroups(name, strict=False)
        if groups is None:
            return None
        groups = dict(groups)  # copy, just in case
        groups['devicename'] = groups.get('devname')
        groups['devalias'] = groups.get('_devalias')
        groups['attributename'] = groups.get('_shortattrname')
        groups['configparam'] = groups.get('fragment')
        return groups

    def getNames(self, fullname, factory=None):
        """Returns a tuple of three elements with  (complete_name, normal_name,
        short_name) or None if no match is found.
        The definitions of each name are:

        - complete: the full URI allowing an unambiguous identification of the
          model within taurus (note: it must include the scheme).
        - normal: an unambiguous URI at the scheme level. Any parts that are
          optional and equal to the scheme's default can be stripped.
          In particular, the scheme name is typically stripped for all schemes.
        - short: a short name (not necessarily a valid URI) useful for display
          in cases where ambiguity is tolerable.

        Example: In a tango system where the default TANGO_HOST is "foo:123"
        and a device "a/b/c" has been defined with alias "bar" and having an
        attribute called "d", getNames would return:

        - for the authority::

            ('tango://foo:123', '//foo:123', 'foo:123')

        - for the device::

            ('tango://foo:123/a/b/c', 'a/b/c', 'bar')

            note: if foo:123 wasn't the default TANGO_HOST, the normal name
            would be '//foo:123/a/b/c'. Equivalent rules apply to Attribute
            normal names.

        - for the attribute::

            ('tango://foo:123/a/b/c/d', 'a/b/c/d', 'd')

            note: if foo123 wasn't the default TANGO_HOST, the normal name
            would be '//foo:123/a/b/c/d'

         - for the attribute (assuming we passed #label)::

            ('tango://foo:123/a/b/c/d#label',
             'a/b/c/d#label',
             'd#label')

         - for the attribute (assuming we did not pass a conf key)::
            ('tango://foo:123/a/b/c/d#',
             'a/b/c/d#',
             'd#')

        Note: it must always be possible to construct the 3 names from a *valid*
        **fullname** URI. If the given URI is valid but it is not the full name,
        it may still be possible in some cases to construct the 3 names, but
        it may involve using defaults provided by the scheme (which may require
        more computation than mere parsing the URI)
        """
        raise NotImplementedError('getNames must be implemented in derived ' +
                                  'classes')


class TaurusAuthorityNameValidator(_TaurusBaseValidator):
    '''Base class for Authority name validators.
    The namePattern will be composed from URI segments as follows:

    <scheme>:<authority>[/<path>][?<query>][#<fragment>]

    Derived classes must provide attributes defining a regexp string for each
    URI segment (they can be empty strings):

    - scheme
    - authority
    - path
    - query
    - fragment
    '''
    pattern = r'^(?P<scheme>%(scheme)s):' + \
              r'(?P<authority>%(authority)s)' + \
              r'((?=/)(?P<path>%(path)s))?' + \
              r'(\?(?P<query>%(query)s))?' + \
              r'(#(?P<fragment>%(fragment)s))?$'

    def getNames(self, name, factory=None):
        '''basic implementation for getNames for authorities. You may
        reimplement it in your scheme if required'''
        groups = self.getUriGroups(name)
        if groups is None:
            return None
        complete = '%(scheme)s:%(authority)s' % groups
        normal = '%(authority)s' % groups
        short = ('%(authority)s' % groups).strip('/')
        return complete, normal, short


class TaurusDeviceNameValidator(_TaurusBaseValidator):
    '''Base class for Device name validators.
    The namePattern will be composed from URI segments as follows:

    <scheme>:[<authority>/]<path>[?<query>][#<fragment>]

    Derived classes must provide attributes defining a regexp string for each
    URI segment (they can be empty strings):

    - scheme
    - authority
    - path
    - query
    - fragment

    Additionally, the namePattern resulting from composing the above segments
    must contain a named group called "devname" (normally within the
    path segment).
    '''
    pattern = r'^(?P<scheme>%(scheme)s):' + \
              r'((?P<authority>%(authority)s)($|(?=[/#?])))?' + \
              r'(?P<path>%(path)s)' + \
              r'(\?(?P<query>%(query)s))?' + \
              r'(#(?P<fragment>%(fragment)s))?$'


class TaurusAttributeNameValidator(_TaurusBaseValidator):
    '''Base class for Attribute name validators.
    The namePattern will be composed from URI segments as follows:

    <scheme>:[<authority>/]<path>[?<query>][#<fragment>]

    Derived classes must provide attributes defining a regexp string for each
    URI segment (they can be empty strings):

    - scheme
    - authority
    - path
    - query
    - fragment

    Additionally, the namePattern resulting from composing the above segments
    must contain a named group called "attrname" (normally within the
    path segment).
    '''
    pattern = r'^(?P<scheme>%(scheme)s):' + \
              r'((?P<authority>%(authority)s)($|(?=[/#?])))?' + \
              r'(?P<path>%(path)s)' + \
              r'(\?(?P<query>%(query)s))?' + \
              r'(#(?P<fragment>%(fragment)s))?$'


class TaurusDatabaseNameValidator(TaurusAuthorityNameValidator):
    '''Backwards-compatibility only. Use TaurusAuthorityNameValidator instead'''

    def __init__(self, *args, **kwargs):
        msg = ('%s is deprecated. Use "Authority" instead of "Database"' %
               self.__class__.__name__)
        from taurus import warning
        warning(msg)
        return TaurusAuthorityNameValidator.__init__(self, *args, **kwargs)


if __name__ == '__main__':

    class FooAttributeNameValidator(TaurusAttributeNameValidator):
        scheme = 'foo'
        authority = '[^?#/]+'
        path = '[^?#]+'
        query = '(?!)'
        fragment = '[^?#]*'

    v = FooAttributeNameValidator()
    name = 'foo://bar#label'
    print(v.isValid(name))
    print(v.getUriGroups(name))



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""
itemsmodel Model and view for new CurveItem configuration
"""
from builtins import object
from builtins import bytes
#raise UnimplementedError('Under Construction!')

import copy

from future.utils import string_types

from taurus.external.qt import Qt
import taurus
from taurus.core.taurushelper import getSchemeFromName
from taurus.core.taurusbasetypes import TaurusElementType
from taurus.core.taurusexception import TaurusException
from taurus.qt.qtcore.mimetypes import TAURUS_MODEL_LIST_MIME_TYPE
from taurus.qt.qtcore.mimetypes import TAURUS_ATTR_MIME_TYPE
from taurus.qt.qtcore.mimetypes import TAURUS_MODEL_MIME_TYPE
from taurus.qt.qtgui.icon import getElementTypeIcon


__all__ = ['TaurusModelModel', 'TaurusModelItem', 'TaurusModelList']

# set some named constants
SRC_ROLE = Qt.Qt.UserRole + 1


class TaurusModelItem(object):
    '''An item object for :class:`TaurusModelModel`. Exposes `display` `icon` and `ok`
    attributes which are calculated and kept in synch with the property `src`'''

    def __init__(self, src=None, display=None):
        self.icon = Qt.QIcon()
        self.ok = True
        self._src = None
        self.setSrc(src)
        if display is not None:
            self.display = display

    def __deepcopy__(self, memo):
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        result.icon = Qt.QIcon(self.icon)
        result.ok = copy.deepcopy(self.ok, memo)
        result._src = copy.deepcopy(self._src, memo)
        result.display = copy.deepcopy(self.display, memo)
        return result

    def __repr__(self):
        ret = "TaurusModelItem('%s')" % (self.display)
        return ret

    def getSrc(self):
        '''getter for src.'''
        return self._src

    def setSrc(self, src):
        '''processes the src and sets the values of _src, display, icon and ok attributes'''
        if src is None:
            self._src, self.display, self.icon, self.ok = '', '(Empty)', Qt.QIcon(
            ), True
            return
        src = str(src).strip()
        if src == self._src:
            return
        self._src = src
        # empty
        if src == '':
            self.display, self.icon, self.ok = '(Empty)', Qt.QIcon(), True
            return
        # TODO tango-centric. For tango devices
        try:
            dev = taurus.Device(src)
            if dev.getDeviceProxy() is not None:
                self.display, self.icon, self.ok = dev.getSimpleName(
                ), getElementTypeIcon(TaurusElementType.Device), True
                return
            else:
                self.display, self.icon, self.ok = src, Qt.QIcon.fromTheme(
                    'network-error'), False
                return
        except:
            # for tango attributes
            try:
                attr = taurus.Attribute(src)
                dev = attr.getParentObj()
            except TaurusException:
                self.display, self.icon, self.ok = src, Qt.QIcon.fromTheme(
                    'dialog-warning'), False
                return
            except Exception:  # @todo: this catchall except is here as an emergency bugfix, but should probably be narrowed to PyTango DevFailed.
                self.display, self.icon, self.ok = src, Qt.QIcon.fromTheme(
                    'network-error'), False
                return
            if getSchemeFromName(attr.getFullName()) == "tango" and \
                            dev.getDeviceProxy() is None:
                    self.display, self.icon, self.ok = src, Qt.QIcon.fromTheme(
                        'network-error'), False
                    return

            self.display, self.icon, self.ok = attr.getSimpleName(
            ), getElementTypeIcon(TaurusElementType.Attribute), True

    # properties
    src = property(getSrc, setSrc)


class TaurusModelModel(Qt.QAbstractListModel):
    ''' A Qt data model for describing taurus models
    '''

    def __init__(self, items=None):
        if items is None:
            items = []
        super(TaurusModelModel, self).__init__()
        self.items = items

    def addItem(self, item):
        '''appends an item to the internal list

        :param item: (TaurusModelItem) '''
        self.items.append(item)

    def dumpData(self):
        '''returns a deep copy of the internal item list representation'''
        return copy.deepcopy(self.items)

    def rowCount(self, index=Qt.QModelIndex()):
        '''reimplemented from :class:`Qt.QAbstractListModel`'''
        return len(self.items)

    def data(self, index, role=Qt.Qt.DisplayRole):
        '''reimplemented from :class:`Qt.QAbstractListModel`'''
        if not index.isValid() or not (0 <= index.row() < self.rowCount()):
            return None
        row = index.row()
        # Display Role
        if role == Qt.Qt.DisplayRole:
            return str(self.items[row].display)
        elif role == Qt.Qt.DecorationRole:
            return self.items[row].icon
        elif role == Qt.Qt.TextColorRole:
            if not self.items[row].src:
                return Qt.QColor('gray')
            return Qt.QColor(self.items[row].ok and 'green' or 'red')
        elif role == SRC_ROLE:
            return str(self.items[row].src)
        elif role == Qt.Qt.ToolTipRole:
            return str(self.items[row].src)
        if role == Qt.Qt.EditRole:
            return str(self.items[row].src)
        return None

    def flags(self, index):
        '''reimplemented from :class:`Qt.QAbstractListModel`'''
        if not index.isValid():
            return Qt.Qt.ItemIsEnabled | Qt.Qt.ItemIsDropEnabled
        return Qt.Qt.ItemFlags(Qt.Qt.ItemIsEnabled | Qt.Qt.ItemIsEditable | Qt.Qt.ItemIsDragEnabled | Qt.Qt.ItemIsSelectable)

    def setData(self, index, value=None, role=Qt.Qt.EditRole):
        '''reimplemented from :class:`Qt.QAbstractListModel`'''
        if index.isValid() and (0 <= index.row() < self.rowCount()):
            row = index.row()
            item = self.items[row]
            if role == Qt.Qt.EditRole:
                item.src = value
            elif role == Qt.Qt.DisplayRole:
                item.display = value
            self.dataChanged.emit(index, index)
            return True
        return False

    def insertRows(self, position=None, rows=1, parentindex=None, items=None):
        '''reimplemented from :class:`Qt.QAbstractListModel`'''
        if position is None or position == -1:
            position = self.rowCount()
        if parentindex is None:
            parentindex = Qt.QModelIndex()
        if items is None:
            slice = [TaurusModelItem() for i in range(rows)]
        else:
            slice = list(items)
            # note that the rows parameter is ignored if items is passed
            rows = len(slice)
        self.beginInsertRows(parentindex, position, position + rows - 1)
        self.items = self.items[:position] + slice + self.items[position:]
        self.endInsertRows()
        return True

    def removeRows(self, position, rows=1, parentindex=None):
        '''reimplemented from :class:`Qt.QAbstractListModel`'''
        if parentindex is None:
            parentindex = Qt.QModelIndex()
        self.beginResetModel()
        self.beginRemoveRows(parentindex, position, position + rows - 1)
        self.items = self.items[:position] + self.items[position + rows:]
        self.endRemoveRows()
        self.endResetModel()
        return True

    def clearAll(self):
        '''clears all rows'''
        self.removeRows(0, self.rowCount())

    def swapItems(self, index1, index2):
        '''swap the items described by index1 and index2 in the list'''
        r1, r2 = index1.row(), index2.row()
        self.items[r1], self.items[r2] = self.items[r2], self.items[r1]
        self.dataChanged.emit(index1, index2)

    def mimeTypes(self):
        '''reimplemented from :class:`Qt.QAbstractListModel`'''
        result = list(Qt.QAbstractItemModel.mimeTypes(self))
        result += [TAURUS_ATTR_MIME_TYPE, TAURUS_MODEL_MIME_TYPE,
                   TAURUS_MODEL_LIST_MIME_TYPE, 'text/plain']
        return result

    def dropMimeData(self, data, action, row, column, parent):
        '''reimplemented from :class:`Qt.QAbstractListModel`'''
        if row == -1 and parent.isValid():
            row = parent.row()
        if data.hasFormat(TAURUS_ATTR_MIME_TYPE):
            items = [bytes(data.data(TAURUS_ATTR_MIME_TYPE)).decode("utf-8")]
        elif data.hasFormat(TAURUS_MODEL_MIME_TYPE):
            items = [bytes(data.data(TAURUS_MODEL_MIME_TYPE)).decode("utf-8")]
        elif data.hasFormat(TAURUS_MODEL_LIST_MIME_TYPE):
            items = bytes(
                data.data(TAURUS_MODEL_LIST_MIME_TYPE)).decode("utf-8").split()
        elif data.hasText():
            items = [str(data.text())]
        else:
            return False
        self.insertItems(row, items)
        return True

    def insertItems(self, row, items):
        '''convenience method to add new rows by passing a list of strings ()

        :param row: (int) the row of the list at which the item insertion
                         starts, if row==-1, items will be appended to the list
        :param items: (seq) a sequence items to add to the list. The objects
                      in the sequence can be either strings, :class:`TaurusModelItem` objects
                      or tuples of valid arguments for initializing :class:`TaurusModelItem` objects
        '''
        itemobjs = []
        for e in items:
            if isinstance(e, TaurusModelItem):
                itemobjs.append(e)
            elif isinstance(e, string_types):
                itemobjs.append(TaurusModelItem(src=e))
            else:  # assuming it is a sequence of arguments that can be passed to the constructor of TaurusModelItem
                itemobjs.append(TaurusModelItem(*e))
        self.insertRows(position=row, items=itemobjs)

    def mimeData(self, indexes):
        '''reimplemented from :class:`Qt.QAbstractListModel`'''
        mimedata = Qt.QAbstractListModel.mimeData(self, indexes)
        if len(indexes) == 1:
            # mimedata.setData(TAURUS_ATTR_MIME_TYPE, self.data(indexes[0]))
            txt = self.data(indexes[0], role=SRC_ROLE)
            mimedata.setText(txt)
        return mimedata
        # mimedata.setData()


class TaurusModelList(Qt.QListView):
    '''A list view widget to display and manage a list of models

    Tries to identify the type of model and show the state of the device/attr
    associated with it. It also allows drag and drop of models and sorting.
    '''

    dataChangedSignal = Qt.pyqtSignal(list, name="dataChanged")

    def __init__(self, parent=None, items=None, designMode=False):
        super(TaurusModelList, self).__init__(parent)
        if items is None:
            items = []
        self._model = TaurusModelModel(items)
        self.setModel(self._model)
        self.setDragDropMode(self.DragDrop)
        # self.setAcceptDrops(True)
        self.setSelectionMode(self.ExtendedSelection)

        self._contextMenu = Qt.QMenu(self)
        self.addRowAction = self._contextMenu.addAction(Qt.QIcon.fromTheme(
            'list-add'), 'Add new row', self.newRow, Qt.QKeySequence.New)
        self.removeSelectedAction = self._contextMenu.addAction(
            Qt.QIcon.fromTheme('list-remove'), 'Remove Selected',
            self.removeSelected, Qt.QKeySequence.Delete)
        self.removeAllAction = self._contextMenu.addAction(Qt.QIcon.fromTheme(
            'edit-clear'), 'Clear all', self.clear, Qt.QKeySequence("Ctrl+Del"))
        self.moveUpAction = self._contextMenu.addAction(Qt.QIcon.fromTheme(
            'go-up'), 'Move up in the list', self._onMoveUpAction, Qt.QKeySequence("Alt+Up"))
        self.moveDownAction = self._contextMenu.addAction(Qt.QIcon.fromTheme(
            'go-down'), 'Move down in the list', self._onMoveDownAction, Qt.QKeySequence("Alt+Down"))
        self.editDisplayAction = self._contextMenu.addAction(
            Qt.QIcon("actions:format-text-italic.svg"),
            'Edit the display (leave the source)', self._onEditDisplay,
            Qt.QKeySequence("Alt+D"))

        self.addActions([self.addRowAction, self.removeSelectedAction, self.removeAllAction,
                         self.moveUpAction, self.moveDownAction, self.editDisplayAction])

        # signal connections
        selectionmodel = self.selectionModel()
        selectionmodel.selectionChanged.connect(self._onSelectionChanged)
        self._model.dataChanged.connect(self._onDataChanged)
        self._model.rowsInserted.connect(self._onDataChanged)
        self._model.rowsRemoved.connect(self._onDataChanged)
        self._onSelectionChanged(Qt.QItemSelection(), Qt.QItemSelection())

    def clear(self):
        '''removes all items from the list'''
        self._model.clearAll()

    def _onEditDisplay(self):
        selected = self.selectionModel().selectedIndexes()
        if len(selected) == 1:
            idx = selected[0]
        else:
            return
        value = self._model.data(idx, role=Qt.Qt.DisplayRole)
        src = self._model.data(idx, role=SRC_ROLE)
        value, ok = Qt.QInputDialog.getText(
            self, "Display Value", "Display value for %s?" % src, Qt.QLineEdit.Normal, value)
        if not ok:
            return
        self._model.setData(idx, value, role=Qt.Qt.DisplayRole)

    def _onSelectionChanged(self, selected, deselected):
        '''updates the status of the actions that depend on the selection'''
        selectedIndexes = self.selectionModel().selectedRows()
        self.removeSelectedAction.setEnabled(len(selectedIndexes) > 0)
        self.moveUpAction.setEnabled(
            len(selectedIndexes) == 1 and selectedIndexes[0].row() > 0)
        self.moveDownAction.setEnabled(len(selectedIndexes) == 1 and (
            0 <= selectedIndexes[0].row() < self._model.rowCount() - 1))
        self.editDisplayAction.setEnabled(len(selectedIndexes) > 0)

    def _onDataChanged(self, *args):
        '''emits a signal containing the current data as a list of strings'''
        self.dataChangedSignal.emit(self.getModelItems())

    def contextMenuEvent(self, event):
        '''see :meth:`QWidget.contextMenuEvent`'''
        self._contextMenu.exec_(event.globalPos())
        event.accept()

    def _onMoveUpAction(self):
        '''slot for move up action'''
        selected = self.selectionModel().selectedIndexes()
        if len(selected) != 1:
            return
        i1 = selected[0]
        i2 = self._model.index(i1.row() - 1)
        self._model.swapItems(i1, i2)
        self.selectionModel().select(i2, Qt.QItemSelectionModel.ClearAndSelect)

    def _onMoveDownAction(self):
        '''slot for move down action'''
        selected = self.selectionModel().selectedIndexes()
        if len(selected) != 1:
            return
        i1 = selected[0]
        i2 = self._model.index(i1.row() + 1)
        self._model.swapItems(i1, i2)
        self.selectionModel().select(i2, Qt.QItemSelectionModel.ClearAndSelect)

    def newRow(self, position=None):
        '''adds an empty row *before* the given position

        :param position: (int or None) position at which the new row will be added.
                         If None passed, it will be added at the end.
        '''
        if position is None:
            selected = self.selectionModel().selectedIndexes()
            if len(selected) == 0:
                position = -1
            elif len(selected) == 1:
                position = selected[0].row()
            else:
                return
        self._model.insertItems(position, [''])

    def removeSelected(self):
        '''removes selected items from the list'''
        selected = sorted(
            [idx.row() for idx in self.selectionModel().selectedRows()], reverse=True)
        for row in selected:  # we remove rows starting from the last one
            self._model.removeRows(row)

    def addModels(self, models):
        '''adds models to the list

        :param models: (list<str>) sequence of model names to be added
        '''
        self._model.insertItems(-1, models)

    def getModelItems(self):
        '''returns the model item objects

        :return: (list<TaurusModelItem>)

        .. seealso:: :meth:`getModelList`
        '''
        return self._model.dumpData()

    def getModelList(self):
        '''returns a the model names corresponding to the items in the list

        :return: (list<str>)

        .. seealso:: :meth:`getModelItems`
        '''
        return [str(s.src) for s in self.getModelItems()]

    @classmethod
    def getQtDesignerPluginInfo(cls):
        return {
            'group': 'Taurus Input',
            'icon': 'logos:taurus.png',
            'container': False,
            'module': 'taurus.qt.qtgui.panel'
        }

if __name__ == "__main__":
    from taurus.qt.qtgui.application import TaurusApplication
    import sys
    app = TaurusApplication(cmd_line_parser=None)
    w = TaurusModelList()
    w.addModels(["item%i" % i for i in range(3)] +
                [TaurusModelItem(src='src1', display='d1')] + [('src2', 'd2')])
    w.show()
    sys.exit(app.exec_())



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""This module contains the base TaurusModel class"""

from builtins import object

import weakref
try:
    from collections.abc import Sequence
except ImportError:  # bck-compat py 2.7
    from collections import Sequence

from .util.log import Logger
from .util.event import (CallableRef,
                         BoundMethodWeakref,
                         _BoundMethodWeakrefWithCall)
from .taurusbasetypes import TaurusEventType, MatchLevel
from .taurushelper import Factory

__all__ = ["TaurusModel"]

__docformat__ = "restructuredtext"


class TaurusModel(Logger):

    _factory = None
    RegularEvent = (TaurusEventType.Change,
                    TaurusEventType.Config, TaurusEventType.Periodic)

    def __init__(self, full_name='', parent=None, serializationMode=None):
        v = self.getNameValidator()
        self._full_name, self._norm_name, self._simp_name = v.getNames(
            full_name, self.factory())

        if self._full_name is None and self._norm_name and self._simp_name is None:
            self.trace("invalid name")

        name = self._simp_name or self._norm_name or self._full_name or 'TaurusModel'
        self.call__init__(Logger, name, parent)

        if serializationMode is None:
            s_obj = parent
            if s_obj is None:
                s_obj = self.factory()
            serializationMode = s_obj.getSerializationMode()
        self._serialization_mode = serializationMode

        self._parentObj = parent
        self._listeners = []

    def __str__name__(self, name):
        return '{0}({1})'.format(self.__class__.__name__, name)

    def __str__(self):
        return self.__str__name__(self.getNormalName())

    def __repr__(self):
        return self.__str__name__(self.getFullName())

    def cleanUp(self):
        self.trace("[TaurusModel] cleanUp")
        self._parentObj = None
        self._listeners = None
        Logger.cleanUp(self)

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # API for Factory access
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    @classmethod
    def factory(cls):
        if cls._factory is None:
            cls._factory = Factory(scheme=cls._scheme)
        return cls._factory

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # API for naming
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    @classmethod
    def getTaurusElementType(cls):
        raise NotImplementedError("TaurusModel.getTaurusElementType cannot"
                                  " be called")

    def getFullName(self):
        return self._full_name

    def getNormalName(self):
        return self._norm_name

    def getSimpleName(self):
        return self._simp_name

    @classmethod
    def isValid(cls, *args, **kwargs):
        return cls.getNameValidator().isValid(*args, **kwargs)

    @classmethod
    def buildModelName(cls, parent_model, relative_name):
        raise NotImplementedError(
            "TaurusModel.buildModelName cannot be called")

    @classmethod
    def getNameValidator(cls):
        raise NotImplementedError("TaurusModel.getNameValidator cannot be"
                                  "called")

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # API for hierarchy access
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def getParentObj(self):
        return self._parentObj

    def getChildObj(self, child_name):
        return None  # TODO: consider raising NotImplementedError instead

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # API for serialization
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def setSerializationMode(self, mode):
        """Sets the serialization mode for the system.

        :param mode: (TaurusSerializationMode) the new serialization mode"""
        self._serialization_mode = mode

    def getSerializationMode(self):
        """Gives the serialization operation mode.

        :return: (TaurusSerializationMode) the current serialization mode"""
        return self._serialization_mode

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # API for value access
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def getDisplayDescrObj(self, cache=True):
        """A brief description of the model. Can be used as tooltip, for example"""
        raise NotImplementedError("TaurusModel.getDisplayDescrObj cannot be"
                                  " called")

    def getDisplayName(self, cache=True, complete=True):
        full_name = self.getFullName()
        normal_name = self.getNormalName()
        simple_name = self.getSimpleName()
        if simple_name:
            ret = simple_name
            if complete:
                ret += " (" + normal_name.upper() + ")"
        elif normal_name:
            ret = normal_name.upper()
        else:
            ret = full_name.upper()
        return ret

    def getFragmentObj(self, fragmentName=None):
        """Returns a fragment object of the model. A fragment is computed from
        a model by evaluating the expression `<model>.<fragmentName>`

        For a simple fragmentName (no dots), this is roughly equivalent to
        getattr(self, fragmentName)

        If the fragment cannot be computed, :class:`AttributeError` is raised

        :param fragmentName: (str or None) the returned value will correspond to
                         the given fragmentName. If None is passed the
                         defaultFragmentName will be used instead.

        :return: (obj) the computed fragment of the modelObj.
        """
        if fragmentName is None:
            fragmentName = self.defaultFragmentName
        try:
            return eval('obj.' + fragmentName, {}, {'obj' : self})
        except Exception as e:
            # Note: always raise AttributeError to comply with existing API
            msg = "Cannot get fragment of {!r}.{!s}: Reason: {!r}"
            raise AttributeError(msg.format(self, fragmentName, e))


    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # API for listeners
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def _listenerDied(self, weak_listener):
        if self._listeners is None:
            return
        try:
            self._listeners.remove(weak_listener)
        except Exception as e:
            self.debug("Problem removing listener: %r", e)

    def _getCallableRef(self, listener, cb=None):
        # return weakref.ref(listener, self._listenerDied)
        meth = getattr(listener, 'eventReceived', None)
        if meth is not None and hasattr(meth, '__call__'):
            return weakref.ref(listener, cb)
        else:
            return CallableRef(listener, cb)

    def addListener(self, listener):
        if self._listeners is None or listener is None:
            return False

        # TODO: _BoundMethodWeakrefWithCall is used as workaround for
        # PyTango #185 issue
        weak_listener = self._getCallableRef(
            listener, _BoundMethodWeakrefWithCall(self._listenerDied))
            # listener, self._listenerDied)
        if weak_listener in self._listeners:
            return False
        self._listeners.append(weak_listener)
        return True

    def removeListener(self, listener):
        if self._listeners is None:
            return
        weak_listener = self._getCallableRef(listener)
        try:
            self._listeners.remove(weak_listener)
        except Exception as e:
            return False
        return True

    def forceListening(self):
        class __DummyListener(object):

            def eventReceived(self, *args):
                pass
        if not hasattr(self, '__dummyListener') or self.__dummyListener is None:
            self.__dummyListener = __DummyListener()
            self.addListener(self.__dummyListener)

    def unforceListening(self):
        if hasattr(self, '__dummyListener') and self.__dummyListener is not None:
            self.removeListener(self.__dummyListener)
            self.__dummyListener = None

    def deleteListener(self, listener):
        self.deprecated("Use removeListener(listener) instead")
        self.removeListener(listener)

    def hasListeners(self):
        """ returns True if anybody is listening to events from this attribute """
        if self._listeners is None:
            return False
        return len(self._listeners) > 0

    def fireEvent(self, event_type, event_value, listeners=None):
        """sends an event to all listeners or a specific one"""

        if listeners is None:
            listeners = self._listeners

        if listeners is None:
            return

        if not isinstance(listeners, Sequence):
            listeners = listeners,

        for listener in listeners:
            if isinstance(listener, weakref.ref) or isinstance(listener, BoundMethodWeakref):
                l = listener()
            else:
                l = listener
            if l is None:
                continue
            meth = getattr(l, 'eventReceived', None)
            if meth is not None and hasattr(meth, '__call__'):
                l.eventReceived(self, event_type, event_value)
            elif hasattr(l, '__call__'):
                l(self, event_type, event_value)

    def isWritable(self):
        return False

    @property
    def name(self):
        return self._simp_name

    @property
    def fullname(self):
        return self._full_name

    parentObj = property(fget=getParentObj)



#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

import pkg_resources
import click
import taurus

import taurus.cli.common
from . import common


@click.group("taurus")
@common.log_level
@common.poll_period
@common.serial_mode
@common.default_formatter
@click.option(
    "--rconsole",
    "rconsole_port",
    type=click.INT,
    metavar="PORT",
    default=None,
    help="Enable remote debugging with rfoo on the given PORT",
)
@click.version_option(version=taurus.Release.version)
def taurus_cmd(
    log_level,
    polling_period,
    serialization_mode,
    default_formatter,
    rconsole_port,
):
    """The main taurus command"""

    # set log level
    taurus.setLogLevel(getattr(taurus, log_level))

    # set polling period
    if polling_period is not None:
        taurus.changeDefaultPollingPeriod(polling_period)

    # set serialization mode
    if serialization_mode is not None:
        from taurus.core.taurusbasetypes import TaurusSerializationMode

        m = getattr(TaurusSerializationMode, serialization_mode)
        taurus.Manager().setSerializationMode(m)

    # enable the remote console port
    if rconsole_port is not None:
        try:
            import rfoo.utils.rconsole

            rfoo.utils.rconsole.spawn_server(port=rconsole_port)
            taurus.info(
                (
                    "rconsole started. "
                    + "You can connect to it by typing: rconsole -p %d"
                ),
                rconsole_port,
            )
        except Exception as e:
            taurus.warning("Cannot spawn debugger. Reason: %s", e)

    # set the default formatter
    if default_formatter is not None:
        from taurus import tauruscustomsettings

        setattr(tauruscustomsettings, "DEFAULT_FORMATTER", default_formatter)


def register_subcommands():
    """Discover and add subcommands to taurus_cmd"""

    # Add subcommands from the taurus_subcommands entry point
    for ep in pkg_resources.iter_entry_points("taurus.cli.subcommands"):
        try:
            subcommand = ep.load()
            taurus_cmd.add_command(subcommand)
        except Exception as e:
            # -----------------------------------------------------------
            # Use info instead of warning in case of nonimportable Qwt5
            # (e.g. when using py3 or Qt5) to avoid spam
            # This special case can be removed when taurus.qt.qtgui.qwt5
            # is moved to a separate plugin, since the entry point will
            # be registered only if the plugin is installed
            if ep.name == "qwt5":
                taurus.info(
                    'Cannot add "%s" subcommand to taurus. Reason: %r',
                    ep.name,
                    e,
                )
                continue
            # -----------------------------------------------------------
            taurus.warning(
                'Cannot add "%s" subcommand to taurus. Reason: %r', ep.name, e
            )


def main():
    """Register subcommands and run taurus_cmd"""

    # set the log level to WARNING avoid spamming the CLI while loading
    # subcommands
    # it will be restored to the desired one first thing in taurus_cmd()
    taurus.setLogLevel(taurus.Warning)

    # register the subcommands
    register_subcommands()

    # launch the taurus command
    taurus_cmd()


if __name__ == "__main__":
    main()



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

""""""

from __future__ import print_function

__all__ = ["QTaurusBaseListener", "QObjectTaurusListener"]

__docformat__ = 'restructuredtext'

from taurus.qt.qtcore.util import baseSignal
from taurus.core.util.log import deprecation_decorator
from taurus.core.tauruslistener import TaurusListener
from taurus.external.qt import Qt


class QTaurusBaseListener(TaurusListener):
    """Base class for QObjects listening to taurus events.

    .. note::
           :meth:`getSignaller` is now unused and deprecated. This is because
           `taurusEvent` is implemented using :func:`baseSignal`, that doesn't
           require the class to inherit from QObject.
    """

    taurusEvent = baseSignal('taurusEvent', object, object, object)

    def __init__(self, name=None, parent=None):
        if name is None:
            name = self.__class__.__name__
        super(QTaurusBaseListener, self).__init__(name, parent=parent)
        self._eventFilters = []
        self.taurusEvent.connect(self.filterEvent)

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # Event handling chain
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def eventReceived(self, evt_src, evt_type, evt_value):
        """The basic implementation of the event handling chain is as
        follows:

            - eventReceived just calls :meth:`fireEvent` which emits a "taurusEvent"
              PyQt signal that is connected (by :meth:`preAttach`) to the
              :meth:`filterEvent` method.
            - After filtering, :meth:`handleEvent` is invoked with the resulting
              filtered event

        .. note::
            in the earlier steps of the chain (i.e., in :meth:`eventReceived`/:meth:`fireEvent`),
            the code is executed in a Python thread, while from eventFilter
            ahead, the code is executed in a Qt thread.
            When writing widgets, one should normally work on the Qt thread
            (i.e. reimplementing :meth:`handleEvent`)

        :param evt_src: (object) object that triggered the event
        :param evt_type: (taurus.core.taurusbasetypes.TaurusEventType) type of event
        :param evt_value: (object) event value
        """
        self.fireEvent(evt_src, evt_type, evt_value)

    def fireEvent(self, evt_src=None, evt_type=None, evt_value=None):
        """Emits a "taurusEvent" signal.
        It is unlikely that you may need to reimplement this method in subclasses.
        Consider reimplementing :meth:`eventReceived` or :meth:`handleEvent`
        instead depending on whether you need to execute code in the python
        or Qt threads, respectively

        :param evt_src: (object or None) object that triggered the event
        :param evt_type: (taurus.core.taurusbasetypes.TaurusEventType or None) type of event
        :param evt_value: (object or None) event value
        """
        try:
            self.taurusEvent.emit(evt_src, evt_type, evt_value)
        except:
            pass

    def filterEvent(self, evt_src=-1, evt_type=-1, evt_value=-1):
        """The event is processed by each and all filters in strict order
        unless one of them returns None (in which case the event is discarded)

        :param evt_src: (object) object that triggered the event
        :param evt_type: (taurus.core.taurusbasetypes.TaurusEventType) type of event
        :param evt_value: (object) event value
        """
        r = evt_src, evt_type, evt_value

        if r == (-1, -1, -1):
            # @todo In an ideal world the signature of this method should be
            # (evt_src, evt_type, evt_value). However there's a bug in PyQt:
            # If a signal is disconnected between the moment it is emitted and
            # the moment the slot is called, then the slot is called without
            # parameters (!?). We added this default values to detect if
            # this is the case without printing an error message each time.
            # If this gets fixed, we should remove this line.
            return

        for f in self._eventFilters:
            r = f(*r)
            if r is None:
                return
        self.handleEvent(*r)

    def handleEvent(self, evt_src, evt_type, evt_value):
        """Event handling. Default implementation does nothing.
        Reimplement as necessary

        :param evt_src: (object or None) object that triggered the event
        :param evt_type: (taurus.core.taurusbasetypes.TaurusEventType or None) type of event
        :param evt_value: (object or None) event value
        """
        pass

    def setEventFilters(self, filters=None):
        """sets the taurus event filters list.
        The filters are run in order, using each output to feed the next filter.
        A filter must be a function that accepts 3 arguments ``(evt_src, evt_type, evt_value)``
        If the event is to be ignored, the filter must return None.
        If the event is  not to be ignored, filter must return a
        ``(evt_src, evt_type, evt_value)`` tuple which may (or not) differ from the input.

        For a library of common filters, see taurus/core/util/eventfilters.py

        :param filters: (sequence) a sequence of filters

        See also: insertEventFilter
        """
        if filters is None:
            filters = []
        self._eventFilters = list(filters)

    def getEventFilters(self):
        """Returns the list of event filters for this widget

        :return: (sequence<callable>) the event filters
        """
        return self._eventFilters

    def insertEventFilter(self, filter, index=-1):
        """insert a filter in a given position

        :param filter: (callable(evt_src, evt_type, evt_value)) a filter
        :param index: (int) index to place the filter (default = -1 meaning place at the end)

        See also: setEventFilters
        """
        self._eventFilters.insert(index, filter)

    @deprecation_decorator(rel='4.0')
    def getSignaller(self):
        return self


class QObjectTaurusListener(Qt.QObject, QTaurusBaseListener):

    def __init__(self, name=None, parent=None):
        self.call__init__wo_kw(Qt.QObject, parent)
        self.call__init__(QTaurusBaseListener, name=name)


class ListenerDemo(QObjectTaurusListener):

    def eventReceived(self, evt_src, evt_type, evt_value):
        self.info("New %s event from %s",
                  taurus.core.taurusbasetypes.TaurusEventType[evt_type], evt_src)
        return super(ListenerDemo, self).eventReceived(evt_src, evt_type, evt_value)

    def handleEvent(self, evt_src, evt_type, evt_value):
        self.info("New %s event from %s",
                  taurus.core.taurusbasetypes.TaurusEventType[evt_type], evt_src)


if __name__ == "__main__":
    import time
    import taurus
    qlistener = ListenerDemo(name="DemoObject")

    attr = taurus.Attribute("sys/tg_test/1/double_scalar")
    attr.addListener(qlistener)

    try:
        while True:
            time.sleep(1)
    except KeyboardInterrupt:
        print("Finished!")



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""a list of helper methods"""

from __future__ import print_function

from builtins import str
from future.utils import string_types
import re
from taurus import tauruscustomsettings
from .util.log import taurus4_deprecation
import click

__all__ = ['check_dependencies', 'log_dependencies', 'getSchemeFromName',
           'getValidTypesForName', 'isValidName', 'makeSchemeExplicit',
           'Manager', 'Factory', 'Device', 'Attribute', 'Configuration',
           'Database', 'Authority', 'Object', 'Logger',
           'Critical', 'Error', 'Warning', 'Info', 'Debug', 'Trace',
           'setLogLevel', 'setLogFormat', 'getLogLevel', 'getLogFormat',
           'resetLogLevel', 'resetLogFormat',
           'enableLogOutput', 'disableLogOutput',
           'log', 'trace', 'debug', 'info', 'warning', 'error', 'fatal',
           'critical', 'deprecated', 'changeDefaultPollingPeriod',
           'getValidatorFromName']

__docformat__ = "restructuredtext"

# regexp for finding the scheme
__SCHEME_RE = re.compile(r'([^:/?#]+):.*')

def __check_pyqt(req='4.8'):
    import PyQt4.Qt
    from pkg_resources import parse_version
    assert(parse_version(PyQt4.Qt.PYQT_VERSION_STR) >= parse_version(req))


def __check_pyqwt5(req='5.2'):
    import PyQt4.Qwt5
    from pkg_resources import parse_version
    assert(parse_version(PyQt4.Qwt5.QWT_VERSION_STR) >= parse_version(req))


def check_dependencies():
    """ 
    Prints a check-list of requirements and marks those that are fulfilled
    """

    # non_pypi is a dictionary with extra:req_list and req_list is a list of
    # (reqname, check) tuples where reqname is the name of a requirement and
    # check is a function that raises an exception if the requirement is not
    # fulfilled
    non_pypi = {'taurus-qt': [('PyQt4>=4.8', __check_pyqt),
                              ('PyQt4.Qwt5>=5.2', __check_pyqwt5),
                              ]
                }
    import pkg_resources
    d = pkg_resources.get_distribution('taurus')
    print("Dependencies for %s:" % d)
    # minimum requirements (without extras)
    for r in d.requires():
        try:
            pkg_resources.require(str(r))
            print('\t[*]', end=' ')
        except Exception:
            print('\t[ ]', end=' ')
        print('%s' % r)
    # requirements for the extras
    print('\nExtras:')
    for extra in sorted(d.extras):
        print("Dependencies for taurus[%s]:" % extra)
        # requirements from PyPI
        for r in d.requires(extras=[extra]):
            try:
                r = str(r).split(';')[0]  # remove marker if present (see #612)
                pkg_resources.require(r)
                print('\t[*]', end=' ')
            except Exception:
                print('\t[ ]', end=' ')
            print('%s' % r)
        # requirements outside PyPI
        for r, check in non_pypi.get(extra, ()):
            try:
                check()
                print('\t[*]', end=' ')
            except Exception:
                print('\t[ ]', end=' ')
            print('%s (not in PyPI)' % r)


def log_dependencies():
    """deprecated since '4.0.4'"""
    from taurus import deprecated
    deprecated(dep='taurus.log_dependencies', rel='4.0.4')


def getSchemeFromName(name, implicit=True):
    """Return the scheme from a taurus name.

    :param name: (str) taurus model name URI.
    :param implicit: (bool) controls whether to return the default scheme
                     (if implicit is True -default-) or None (if implicit is
                     False) in case `model` does not contain the scheme name
                     explicitly. The default scheme may be defined in
                     :ref:`tauruscustomsettings` ('tango' is assumed if
                     not defined)
    """
    m = __SCHEME_RE.match(name)
    if m is not None:
        return m.groups()[0]
    if implicit:
        return getattr(tauruscustomsettings, 'DEFAULT_SCHEME', "tango")
    else:
        return None


def getValidatorFromName(name):
    """Helper for obtaining the validator object corresponding to the
    given name.

    :return: model name validator or None if name is not a supported model name
    """

    try:
        factory = Factory(scheme=getSchemeFromName(name))
    except:
        return None
    return factory.getValidatorFromName(name)



def makeSchemeExplicit(name, default=None):
    """return the name guaranteeing that the scheme is present. If name already
    contains the scheme, it is returned unchanged.

    :param name: (str) taurus model name URI.
    :param default: (str) The default scheme to use. If no default is passed,
                     the one defined in tauruscustomsettings.DEFAULT_SCHEME is
                     used.

    :return: the name with the explicit scheme.
    """
    if getSchemeFromName(name, implicit=False) is None:
        if default is None:
            default = getattr(tauruscustomsettings, 'DEFAULT_SCHEME', "tango")
        return "%s:%s" % (default, name)
    else:
        return name


def getValidTypesForName(name, strict=None):
    """
    Returns a list of all Taurus element types for which `name` is a valid
    model name (while in many cases a name may only be valid for one
    element type, this is not necessarily true in general)

    :param name: (str) taurus model name
    :param strict: (bool) If True, names that are not RFC3986-compliant but
                   which would be accepted for backwards compatibility are
                   considered valid.

    :return: (list<TaurusElementType.element>) where element can be one of:
             `Attribute`, `Device` or `Authority`
    """
    try:
        factory = Factory(scheme=getSchemeFromName(name))
    except:
        return []
    return factory.getValidTypesForName(name, strict=strict)


def isValidName(name, etypes=None, strict=None):
    """Returns True is the given name is a valid Taurus model name. If
    `etypes` is passed, it returns True only if name is valid for at least
    one of the given the element types. Otherwise it returns False.
    For example::

        isValidName('tango:foo')--> True
        isValidName('tango:a/b/c', [TaurusElementType.Attribute]) --> False

    :param name: (str) the string to be checked for validity
    :param etypes: (seq<TaurusElementType>) if given, names will only be
                   considered valid if they represent one of the given
                   element types. Supported element types are:
                   `Attribute`, `Device` and `Authority`
    :param strict: (bool) If True, names that are not RFC3986-compliant but
                   which would be accepted for backwards compatibility are
                   considered valid.

    :return: (bool)
    """
    validtypes = getValidTypesForName(name, strict=strict)
    if etypes is None:
        return bool(validtypes)
    for e in etypes:
        if e in validtypes:
            return True
    return False


def Manager():
    """Returns the one and only TaurusManager

    It is a shortcut to::

        import taurus.core
        manager = taurus.core.taurusmanager.TaurusManager()

    :return: the TaurusManager
    :rtype: :class:`taurus.core.taurusmanager.TaurusManager`

    .. seealso:: :class:`taurus.core.taurusmanager.TaurusManager`
    """
    from taurus.core.taurusmanager import TaurusManager
    return TaurusManager()


def Factory(scheme=None):
    """Returns the one and only Factory for the given scheme

    It is a shortcut to::

        import taurus.core.taurusmanager
        manager = taurus.core.taurusmanager.TaurusManager()
        factory = manager.getFactory(scheme)

    :param scheme: a string representing the scheme. Default value is None meaning ``tango`` scheme
    :type scheme: str
    :return: a taurus factory
    :rtype: :class:`taurus.core.taurusfactory.TaurusFactory`
    """
    manager = Manager()
    f = manager.getFactory(scheme=scheme)
    if f is None:
        from taurus.core.taurusexception import TaurusException
        if scheme is None:
            scheme = "default scheme '" + manager.default_scheme + "'"
        else:
            scheme = "'" + scheme + "'"
        raise TaurusException('Cannot create Factory for %s' % scheme)
    return f()


def Device(device_name):
    """Returns the taurus device for the given device name

    It is a shortcut to::

        import taurus.core.taurusmanager
        manager = taurus.core.taurusmanager.TaurusManager()
        factory = manager.getFactory()
        device  = factory.getDevice(device_name)

    :param device_name: the device name
    :type device_name: str
    :return: a taurus device
    :rtype: :class:`taurus.core.taurusdevice.TaurusDevice`
    """
    return Factory(scheme=getSchemeFromName(device_name)).getDevice(device_name)


def Attribute(dev_or_attr_name, attr_name=None):
    """Returns the taurus attribute for either the pair (device name, attribute name)
    or full attribute name

    - Attribute(full_attribute_name)
    - Attribute(device_name, attribute_name)

    It is a shortcut to::

        import taurus.core.taurusmanager
        manager = taurus.core.taurusmanager.TaurusManager()
        factory = manager.getFactory()
        attribute  = factory.getAttribute(full_attribute_name)

    or::

        import taurus.core.taurusmanager
        manager = taurus.core.taurusmanager.TaurusManager()
        factory = manager.getFactory()
        device  = factory.getDevice(device_name)
        attribute = device.getAttribute(attribute_name)

    :param dev_or_attr_name: the device name or full attribute name
    :type dev_or_attr_name: str or TaurusDevice
    :param attr_name: attribute name
    :type attr_name: str
    :return: a taurus attribute
    :rtype: :class:`taurus.core.taurusattribute.TaurusAttribute`
    """
    import types

    if attr_name is None:
        return Factory(scheme=getSchemeFromName(dev_or_attr_name)).getAttribute(dev_or_attr_name)
    else:
        if isinstance(dev_or_attr_name, string_types):
            dev = Device(dev_or_attr_name)
        else:
            dev = dev_or_attr_name
        return dev.getAttribute(attr_name)


@taurus4_deprecation(alt='Attribute')
def Configuration(attr_or_conf_name, conf_name=None):
    """Returns the taurus configuration for either the pair
    (attribute name, conf name) or full conf name

    - Configuration(full_conf_name)
    - Configuration(attribute_name, conf_name)

    It is a shortcut to::

        import taurus.core.taurusmanager
        manager = taurus.core.taurusmanager.TaurusManager()
        factory = manager.getFactory()
        conf  = factory.getConfiguration(attr_or_conf_name)

    or::

        import taurus.core.taurusmanager
        manager = taurus.core.taurusmanager.TaurusManager()
        factory = manager.getFactory()
        attribute  = factory.getAttribute(attribute_name)
        conf = attribute.getConfig(conf_name)

    :param attr_or_conf_name: the full attribute name or full conf name
    :type attr_or_conf_name: str
    :param conf_name: conf name
    :type conf_name: str or None
    :return: a taurus configuration
    :rtype: :class:`taurus.core.taurusconfiguration.TaurusConfiguration`
    """
    return Attribute(attr_or_conf_name)


@taurus4_deprecation(alt='Authority')
def Database(name=None):
    return Authority(name=name)


def Authority(name=None):
    """Returns a taurus authority

    It is a shortcut to::

        import taurus.core.taurusmanager
        manager = taurus.core.taurusmanager.TaurusManager()
        factory = manager.getFactory()
        db  = factory.getAuthority(dname)

    :param name: authority name. If None (default) it will return the default
                 authority of the default scheme. For example, if the default
                 scheme is tango, it will return the default TANGO_HOST database
    :type name: str or None
    :return: a taurus authority
    :rtype: :class:`taurus.core.taurusauthority.TaurusAuthority`
    """
    return Factory(getSchemeFromName(name or '')).getAuthority(name)


def Object(*args):
    """Returns an taurus object of given class for the given name

    Can be called as:

      - Object(name)
      - Object(cls, name)

    Where:

      - `name` is a model name (str)
      - `cls` is a class derived from TaurusModel

    If `cls` is not given, Object() will try to guess it from `name`.

    :return: a taurus object
    :rtype: :class:`taurus.core.taurusmodel.TaurusModel`
    """
    if len(args) == 1:
        klass, name = None, args[0]
    elif len(args) == 2:
        klass, name = args
    else:
        msg = 'Object() takes either 1 or 2 arguments (%i given)' % len(args)
        raise TypeError(msg)
    factory = Factory(getSchemeFromName(name))
    if klass is None:
        klass = factory.findObjectClass(name)
    return factory.getObject(klass, name)


@click.command('check-deps')
def check_dependencies_cmd():
    """
    Shows the taurus dependencies and checks if they are available
    """
    check_dependencies()


from taurus.core.util import log as __log_mod

Logger = __log_mod.Logger
Critical = Logger.Critical
Fatal = Logger.Fatal
Error = Logger.Error
Warning = Logger.Warning
Info = Logger.Info
Debug = Logger.Debug
Trace = Logger.Trace

setLogLevel = Logger.setLogLevel
setLogFormat = Logger.setLogFormat
getLogLevel = Logger.getLogLevel
getLogFormat = Logger.getLogFormat
resetLogLevel = Logger.resetLogLevel
resetLogFormat = Logger.resetLogFormat

enableLogOutput = Logger.enableLogOutput
disableLogOutput = Logger.disableLogOutput

log = __log_mod._log
trace = __log_mod.trace
debug = __log_mod.debug
info = __log_mod.info
warning = __log_mod.warning
error = __log_mod.error
fatal = __log_mod.fatal
critical = __log_mod.critical
deprecated = __log_mod.deprecated


def changeDefaultPollingPeriod(period):
    Manager().changeDefaultPollingPeriod(period)

#del __log_mod
#del __translate_version_str2int



#!/usr/bin/env python

##############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

import sys
from setuptools import setup, find_packages


def get_release_info():
    if sys.version_info >= (3, 5):
        from importlib.util import spec_from_file_location, module_from_spec
        from pathlib import Path
        path = Path(__file__).parent / 'lib' / 'taurus' / 'core' / 'release.py'
        spec = spec_from_file_location('release', path.as_posix())
        module = module_from_spec(spec)
        spec.loader.exec_module(module)
    else:  # for py27
        import os
        import imp
        module_name = "release"
        setup_dir = os.path.dirname(os.path.abspath(__file__))
        release_dir = os.path.join(setup_dir, 'lib', 'taurus', 'core')
        data = imp.find_module(module_name, [release_dir])
        module = imp.load_module(module_name, *data)
    return module


release = get_release_info()

package_dir = {'': 'lib'}

packages = find_packages(where='lib')

provides = [
    'taurus',
]

install_requires = [
    'numpy>=1.1',
    'pint>=0.8',
    'future',
    'click',
    'enum34;python_version<"3.4"',
]

extras_require = {
    'taurus-qt': [# 'PyQt4 >=4.8',
                  # 'PyQt4.Qwt5 >=5.2.0',
                  'ply >=2.3',  # synoptics
                  'lxml >=2.1',  # taurusgui
                  'guiqwt >=3',  # extra_guiqwt
                  ],
    'taurus-tango': ['PyTango >=7.1',
                     ],
    'taurus-epics': ['pyepics >=3.2.4',
                     ],
    'taurus-h5file': ['h5file',
                      ],
    # separate the editor from 'taurus-qt' to avoid forcing spyder>3
    # which implies many more dependencies that may be hard on older system
    'taurus-qt-editor': ['spyder >=3',
                         ],
}

console_scripts = [
    'taurus = taurus.cli:main',
]

taurus_subcommands = [
    'testsuite = taurus.test.testsuite:testsuite_cmd',
    'config = taurus.qt.qtgui.panel.taurusconfigeditor:config_cmd',
    'qwt5 = taurus.qt.qtgui.qwt5.cli:qwt5',
    'device = taurus.qt.qtgui.panel.taurusdevicepanel:device_cmd',
    'panel = taurus.qt.qtgui.panel.taurusdevicepanel:panel_cmd',
    'gui = taurus.qt.qtgui.taurusgui.taurusgui:gui_cmd',
    'newgui = taurus.qt.qtgui.taurusgui.taurusgui:newgui_cmd',
    'designer = taurus.qt.qtdesigner.taurusdesigner:designer_cmd',
    'guiqwt = taurus.qt.qtgui.extra_guiqwt.cli:guiqwt',
    'icons = taurus.qt.qtgui.icon.catalog:icons_cmd',
    'form = taurus.qt.qtgui.panel.taurusform:form_cmd',
    'demo = taurus.qt.qtgui.panel.taurusdemo:demo_cmd',
    'logmon = taurus.core.util.remotelogmonitor:logmon_cmd',
    'qlogmon = taurus.qt.qtgui.table.qlogtable:qlogmon_cmd',
    'check-deps = taurus.core.taurushelper:check_dependencies_cmd',
    'plot = taurus.cli.alt:plot_cmd',
    'trend = taurus.cli.alt:trend_cmd',
    'trend2d = taurus.cli.alt:trend2d_cmd',
    'image = taurus.cli.alt:image_cmd',
]

plot_alternatives = [
    "qwt5 = taurus.qt.qtgui.qwt5:TaurusPlot",
]

trend_alternatives = [
    "qwt5 = taurus.qt.qtgui.qwt5:TaurusTrend",
]

trend2d_alternatives = [
    "guiqwt = taurus.qt.qtgui.extra_guiqwt:TaurusTrend2DDialog",
]

image_alternatives = [
    "guiqwt = taurus.qt.qtgui.extra_guiqwt:TaurusImageDialog",
]

model_selectors = [
    'Tango = taurus.qt.qtgui.panel.taurusmodelchooser:TangoModelSelectorItem',
]

formatters = [
    'taurus = taurus.qt.qtgui.base:defaultFormatter',
    'tango = taurus.core.tango.util:tangoFormatter',
    '{:2.3e} = taurus.qt.qtgui.base:expFormatter',
    '{:.5f} = taurus.qt.qtgui.base:floatFormatter',
]

entry_points = {
    'console_scripts': console_scripts,
    'taurus.cli.subcommands': taurus_subcommands,
    'taurus.model_selector.items': model_selectors,
    'taurus.qt.formatters': formatters,
    'taurus.plot.alts': plot_alternatives,
    'taurus.trend.alts': trend_alternatives,
    'taurus.trend2d.alts': trend2d_alternatives,
    'taurus.image.alts': image_alternatives,
}

classifiers = [
    'Development Status :: 5 - Production/Stable',
    'Environment :: Console',
    'Environment :: X11 Applications :: Qt',
    'Environment :: Win32 (MS Windows)',
    'Intended Audience :: Developers',
    'Intended Audience :: Science/Research',
    'License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)',
    'Natural Language :: English',
    'Operating System :: Microsoft :: Windows',
    'Operating System :: POSIX',
    'Operating System :: POSIX :: Linux',
    'Operating System :: Unix',
    'Operating System :: OS Independent',
    'Programming Language :: Python',
    'Programming Language :: Python :: 2.7',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Topic :: Scientific/Engineering',
    'Topic :: Software Development :: Libraries',
    'Topic :: Software Development :: User Interfaces',
    'Topic :: Software Development :: Widget Sets',
]


setup(name='taurus',
      version=release.version,
      description=release.description,
      long_description=release.long_description,
      author=release.authors['Tiago_et_al'][0],
      maintainer=release.authors['Community'][0],
      maintainer_email=release.authors['Community'][1],
      url=release.url,
      download_url=release.download_url,
      platforms=release.platforms,
      license=release.license,
      keywords=release.keywords,
      packages=packages,
      package_dir=package_dir,
      classifiers=classifiers,
      include_package_data=True,
      entry_points=entry_points,
      provides=provides,
      python_requires='>=2.7',
      install_requires=install_requires,
      extras_require=extras_require,
      test_suite='taurus.test.testsuite.get_taurus_suite',
      )



#!/usr/bin/env python
#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

__all__ = ['EvaluationAttribute']

import numpy
import re
import weakref

from taurus.core.units import Quantity
from taurus.core.taurusattribute import TaurusAttribute
from taurus.core.taurusbasetypes import SubscriptionState, TaurusEventType, \
    TaurusAttrValue, TaurusTimeVal, AttrQuality, DataType
from taurus.core.taurusexception import TaurusException
from taurus.core.taurushelper import Attribute, Manager
from taurus.core import DataFormat
from taurus.core.util.log import debug, taurus4_deprecation

from taurus.core.evaluation.evalvalidator import QUOTED_TEXT_RE, PY_VAR_RE


class EvaluationAttrValue(TaurusAttrValue):
    """Reimplementation of TaurusAttrValue to provide bck-compat via a ref

    """
    # TODO: remove this class once the standard widgets are adapted to TEP14

    def __init__(self, attr=None, config=None):
        # config parameter is kept for backwards compatibility only
        TaurusAttrValue.__init__(self)
        if config is not None:
            from taurus.core.util.log import deprecated
            deprecated(dep='"config" kwarg', alt='"attr"', rel='4.0')
            attr = config
        if attr is None:
            self._attrRef = None
        else:
            self._attrRef = weakref.proxy(attr)
        self.config = self._attrRef

    def __getattr__(self, name):
        # Do not try to delegate special methods
        if name.startswith('__') and name.endswith('__'):
            raise AttributeError("'%s' object has no attribute %s"
                                 % (self.__class__.__name__, name))   
        try:
            ret = getattr(self._attrRef, name)
        except AttributeError:
            raise AttributeError('%s has no attribute %s'
                                 % (self.__class__.__name__, name))
        # return the attr but only after warning
        from taurus.core.util.log import deprecated
        deprecated(dep='EvaluationAttrValue.%s' % name,
                   alt='EvaluationAttribute.%s' % name, rel='4.0')
        return ret

    # --------------------------------------------------------
    # This is for backwards compat with the API of taurus < 4
    #
    @taurus4_deprecation(alt='.rvalue')
    def _get_value(self):
        '''for backwards compat with taurus < 4'''
        debug(repr(self))
        try:
            return self.__fix_int(self.rvalue.magnitude)
        except AttributeError:
            return self.rvalue

    @taurus4_deprecation(alt='.rvalue')
    def _set_value(self, value):
        '''for backwards compat with taurus < 4'''
        debug('Setting %r to %s' % (value, self.name))

        if self.rvalue is None:  # we do not have a previous rvalue
            import numpy
            dtype = numpy.array(value).dtype
            if numpy.issubdtype(dtype, int) or numpy.issubdtype(dtype, float):
                msg = 'Refusing to set ambiguous value (deprecated .value API)'
                raise ValueError(msg)
            else:
                self.rvalue = value
        elif hasattr(self.rvalue, 'units'):  # we do have it and is a Quantity
            self.rvalue = Quantity(value, units=self.rvalue.units)
        else:  # we do have a previous value and is not a quantity
            self.rvalue = value

    value = property(_get_value, _set_value)

    @taurus4_deprecation(alt='.wvalue')
    def _get_w_value(self):
        '''for backwards compat with taurus < 4'''
        debug(repr(self))
        try:
            return self.__fix_int(self.wvalue.magnitude)
        except AttributeError:
            return self.wvalue

    @taurus4_deprecation(alt='.wvalue')
    def _set_w_value(self, value):
        '''for backwards compat with taurus < 4'''
        debug('Setting %r to %s' % (value, self.name))

        if self.wvalue is None:  # we do not have a previous wvalue
            import numpy
            dtype = numpy.array(value).dtype
            if numpy.issubdtype(dtype, int) or numpy.issubdtype(dtype, float):
                msg = 'Refusing to set ambiguous value (deprecated .value API)'
                raise ValueError(msg)
            else:
                self.wvalue = value
        elif hasattr(self.wvalue, 'units'):  # we do have it and is a Quantity
            self.wvalue = Quantity(value, units=self.wvalue.units)
        else:  # we do have a previous value and is not a quantity
            self.wvalue = value

    w_value = property(_get_w_value, _set_w_value)

    @property
    @taurus4_deprecation(alt='.error')
    def has_failed(self):
        return self.error

    def __fix_int(self, value):
        '''cast value to int if  it is an integer.
        Works on scalar and non-scalar values'''
        if self.type != DataType.Integer:
            return value
        try:
            return int(value)
        except TypeError:
            import numpy
            return numpy.array(value, dtype='int')


class EvaluationAttribute(TaurusAttribute):
    """A :class:`TaurusAttribute` that can be used to perform mathematical
    operations involving other arbitrary Taurus attributes. The mathematical
    operation is described in the attribute name itself. An Evaluation Attribute
    will keep references to any other attributes being referenced and it will
    update its own value whenever any of the referenced attributes change.

    .. seealso:: :mod:`taurus.core.evaluation`

    .. warning:: In most cases this class should not be instantiated directly.
                 Instead it should be done via the
                 :meth:`EvaluationFactory.getAttribute`
    """
    # helper class property that stores a reference to the corresponding
    # factory
    _factory = None
    _scheme = 'eval'

    def __init__(self, name='', parent=None, **kwargs):
        self.call__init__(TaurusAttribute, name, parent, **kwargs)
        self._value = EvaluationAttrValue(attr=self)

        self._label = self.getSimpleName()
        self._references = []
        self._validator = self.getNameValidator()
        self._transformation = None
        self.__subscription_state = SubscriptionState.Unsubscribed
        self._value_setter = None

        # This should never be None because the init already ran the validator
        trstring = self._validator.getExpandedExpr(str(name))

        trstring, ok = self.preProcessTransformation(trstring)

        if ok:
            self._transformation = trstring
            self.applyTransformation()

        self._initWritable(trstring)

    def _initWritable(self, trstring):
        # Determine if the device supports writing to this attribute and
        # initialize the writing infrastructure accordingly
        self.writable = False
        try:
            dev = self.getParentObj()
            names = trstring.split('.')
            obj = instance = dev.getSafe()[names[0]]
            for n in names[1:-1]:
                obj = getattr(obj, n)
            obj = getattr(obj.__class__, names[-1])
        except Exception as e:
            # self.info("%r", e)
            return
        ######################################################################
        # This check wrongly returns false for writable properties defined with
        # the @x.setter decorator
        # TODO: Improve this
        self.writable = hasattr(obj, 'fset') and obj.fset is not None
        #######################################################################
        if self.writable:
            def value_setter(value):
                obj.fset(instance, value)

            self._value_setter = value_setter
            self._value.wvalue = self._value.rvalue

    @staticmethod
    def getId(obj, idFormat=r'_V%i_'):
        '''returns an id string for the given object which has the following
           two properties:
            - It is unique for this object during all its life
            - It is a string that can be used as a variable or method name

        :param obj: (object) the python object whose id is requested
        :param idFormat: (str) a format string containing a "`%i`" which,
                         when expanded must be a valid variable name
                         (i.e. it must match
                         `[a-zA-Z_][a-zA-Z0-9_]*`). The default is `_V%i_`
        '''
        return idFormat % id(obj)

    def preProcessTransformation(self, trstring):
        """
        parses the transformation string and creates the necessary symbols for
        the evaluator. It also connects any referenced attributes so that the
        transformation gets re-evaluated if they change.

        :param trstring: (str) a string to be pre-processed

        :return: (tuple<str,bool>) a tuple containing the processed string
                 and a boolean indicating if the preprocessing was successful.
                 if ok==True, the string is ready to be evaluated
        """
        # disconnect previously referenced attributes and clean the list
        for ref in self._references:
            ref.removeListener(self)
        self._references = []

        # get symbols
        evaluator = self.getParentObj()

        # Find references in the string, create references if needed,
        # connect to them and substitute the references by their id
        v = self._validator
        refs = v.getRefs(trstring, ign_quoted=True)
        for r in refs:
            symbol = self.__ref2Id(r)
            trstring = v.replaceUnquotedRef(trstring, '{%s}' % r, symbol)

        # validate the expression (look for missing symbols)
        safesymbols = list(evaluator.getSafe().keys())
        # remove literal text strings from the validation
        trimmedstring = re.sub(QUOTED_TEXT_RE, '', trstring)
        for s in set(re.findall(PY_VAR_RE, trimmedstring)):
            if s not in safesymbols:
                self.warning('Missing symbol "%s"' % s)
                return trstring, False

        # If all went ok, enable/disable polling based on whether
        # there are references or not
        # TODO: Should we activate polling only if explicit in the expression?
        wantpolling = not self.isUsingEvents()
        haspolling = self.isPollingEnabled()
        if wantpolling:
            self._activatePolling()
        elif haspolling and not wantpolling:
            self.disablePolling()

        return trstring, True

    def __ref2Id(self, ref):
        """
        Returns the id of an
        existing taurus attribute corresponding to the match.
        The attribute is created if it didn't previously exist.

        :param ref: (str)  string corresponding to a reference. e.g. eval:1
        """
        refobj = self.__createReference(ref)
        return self.getId(refobj)

    def __createReference(self, ref):
        '''
        Receives a taurus attribute name and creates/retrieves a reference to
        the attribute object. If the object was not already referenced, it adds
        it to the reference list and adds its id and current value to the
        symbols dictionary of the evaluator.

        :param ref: (str)

        :return: (TaurusAttribute)

        '''
        refobj = Attribute(ref)
        if refobj not in self._references:
            evaluator = self.getParentObj()
            v = refobj.read().rvalue
            # add its rvalue to the evaluator symbols
            evaluator.addSafe({self.getId(refobj): v})
            # add the object to the reference list
            self._references.append(refobj)
        return refobj

    def eventReceived(self, evt_src, evt_type, evt_value):
        try:
            v = evt_value.rvalue
        except AttributeError:
            self.trace('Ignoring event from %s' % repr(evt_src))
            return
        # update the corresponding value
        evaluator = self.getParentObj()
        evaluator.addSafe({self.getId(evt_src): v})
        # re-evaluate
        self.applyTransformation()
        # notify listeners that the value changed
        if self.isUsingEvents():
            self.fireEvent(evt_type, self._value)

    def applyTransformation(self):
        if self._transformation is None:
            return
        try:
            evaluator = self.getParentObj()
            rvalue = evaluator.eval(self._transformation)
            # --------------------------------------------------------- 
            # Workaround for https://github.com/hgrecco/pint/issues/509
            # The numpy.shape method over a Quantity mutates
            # the type of its magnitude.
            # TODO: remove "if" when the bug is solved in pint
            if hasattr(rvalue, "magnitude"):
                value_dimension = len(numpy.shape(rvalue.magnitude))
            else:
                value_dimension = len(numpy.shape(rvalue))
            # ---------------------------------------------------------
            value_dformat = DataFormat(value_dimension)
            self.data_format = value_dformat
            self.type = self._encodeType(rvalue, value_dformat)
            if self.type is None:
                raise TypeError("Unsupported returned type, %r" % rvalue)
            if self.type in [DataType.Integer, DataType.Float] and\
                    not isinstance(rvalue, Quantity):
                self.debug("Transformation converted to Quantity")
                rvalue = Quantity(rvalue)
            elif self.type == DataType.Boolean and value_dimension > 1:
                self.debug("Transformation converted to numpy.array")
                rvalue = numpy.array(rvalue)
            self._value.rvalue = rvalue
            self._value.time = TaurusTimeVal.now()
            self._value.quality = AttrQuality.ATTR_VALID
        except Exception as e:
            self._value.quality = AttrQuality.ATTR_INVALID
            msg = " the function '%s' could not be evaluated. Reason: %s" \
                % (self._transformation, repr(e))
            self.warning(msg)

    def _encodeType(self, value, dformat):
        ''' Encode the value type into Taurus data type. In case of non-zero
        dimension attributes e.g. 1D, 2D the type corresponds to the type of the
        first element.

        :param value: (obj)
        :param dformat: (taurus.DataFormat)

        :return: (taurus.DataType)
        '''
        # TODO: Should we fall back to DataType.Object instead of None?
        try:  # handle Quantities
            value = value.magnitude
        except AttributeError:
            pass
        try:  # handle numpy arrays
            value = value.item(0)
        except ValueError:  # for numpy arrays of shape=()
            value = value.item()
        except AttributeError:  # for bool, bytes, str, seq<str>...
            if dformat is DataFormat._1D:
                value = value[0]
            elif dformat is DataFormat._2D:
                value = value[0][0]
        dataType = type(value)
        return DataType.from_python_type(dataType)

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # Necessary to overwrite from TaurusAttribute
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    def isBoolean(self):
        return isinstance(self._value.rvalue, bool)

    def getDisplayValue(self, cache=True):
        return str(self.read(cache=cache).rvalue)

    def encode(self, value):
        return value

    def decode(self, attr_value):
        return attr_value

    def write(self, value, with_read=True):
        if not self.isWritable():
            raise TaurusException('Attempt to write on read-only attribute %s',
                                  self.getFullName())
        self._value_setter(value)
        self._value.wvalue = value

        if with_read:
            ret=self.read(cache=False)
            return ret

    def read(self, cache=True):
        '''returns the value of the attribute.

        :param cache: (bool) If True (default), the last calculated value will
                      be returned. If False, the referenced values will be re-
                      read and the transformation string will be re-evaluated

        :return: attribute value
        '''

        if not cache:
            symbols = {}
            for ref in self._references:
                symbols[self.getId(ref)] = ref.read(cache=False).rvalue
            evaluator = self.getParentObj()
            evaluator.addSafe(symbols)
            self.applyTransformation()
        return self._value

    def poll(self):
        v = self.read(cache=False)
        self.fireEvent(TaurusEventType.Periodic, v)

    def isUsingEvents(self):
        # if this attributes depends from others, then we consider it uses
        # events
        return bool(len(self._references))

#------------------------------------------------------------------------------
    def __fireRegisterEvent(self, listener):
        # fire a first change event
        try:
            v = self.read()
            self.fireEvent(TaurusEventType.Change, v, listener)
        except:
            self.fireEvent(TaurusEventType.Error, None, listener)

    def addListener(self, listener):
        """ Add a TaurusListener object in the listeners list.
            If it is the first listener, it triggers the subscription to
            the referenced attributes.
            If the listener is already registered nothing happens."""
        initial_subscription_state = self.__subscription_state

        ret = TaurusAttribute.addListener(self, listener)

        if not ret:
            return ret

        if self.__subscription_state == SubscriptionState.Unsubscribed:
            for refobj in self._references:
                # subscribe to the referenced attributes
                refobj.addListener(self)
            self.__subscription_state = SubscriptionState.Subscribed

        assert len(self._listeners) >= 1
        # if initial_subscription_state == SubscriptionState.Subscribed:
        if len(self._listeners) > 1 and \
           (initial_subscription_state == SubscriptionState.Subscribed or
                self.isPollingActive()):
            Manager().enqueueJob(self.__fireRegisterEvent,
                                 job_args=((listener,),))
        return ret

    def removeListener(self, listener):
        """ Remove a TaurusListener from the listeners list. If polling enabled
            and it is the last element then stop the polling timer.
            If the listener is not registered nothing happens."""
        ret = TaurusAttribute.removeListener(self, listener)

        if ret and not self.hasListeners():
            self._deactivatePolling()
            self.__subscription_state = SubscriptionState.Unsubscribed
        return ret



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""
taurustrend.py: Generic trend widget for Taurus
"""
__all__ = ["TaurusTrend2DDialog"]

from guiqwt.plot import ImageDialog
from taurus.external.qt import Qt
import taurus.core
from taurus.qt.qtcore.configuration import BaseConfigurableClass
from taurus.qt.qtgui.base import TaurusBaseWidget
from taurus.qt.qtgui.extra_guiqwt.image import TaurusTrend2DItem
from taurus.qt.qtgui.extra_guiqwt.tools import (TaurusModelChooserTool,
                                                TimeAxisTool, AutoScrollTool,
                                                AutoScaleXTool, AutoScaleYTool,
                                                AutoScaleZTool)
import taurus.cli.common


class TaurusTrend2DDialog(ImageDialog, TaurusBaseWidget):
    """
    This is a widget for displaying trends from 1D Taurus attributes (i.e.,
    representing the variation over time of a 1D array). Sometimes this kind of
    plots are also known as "spectrograms".

    The widget shows a 3D plot (Z represented with colors) where the values in
    the 1D array are plotted in the Y-Z plane and are stacked along the X axis.
    """
    _modifiableByUser = True

    def __init__(self, parent=None, designMode=False, toolbar=True,
                 stackMode='deltatime', buffersize=512, options=None,
                 autoscale='xyz', **kwargs):
        """

        :param param: param to be passed to XYImageItem constructor
        :param buffersize: (int) size of the stack
        :param stackMode: (str) can be 'datetime', 'timedelta' or 'event'
        :param autoscale: (str) if autscale string contains 'x', the x axis 
                          will be autoscaled. Same with 'y' and 'z'.
                          Defaults to 'xyz'
        """
        """see :class:`guiqwt.plot.ImageDialog` for other valid initialization
        parameters"""
        defaultOptions = dict(lock_aspect_ratio=False)
        if options is not None:
            defaultOptions.update(options)
        ImageDialog.__init__(self, parent=parent, toolbar=toolbar,
                             options=defaultOptions, **kwargs)
        TaurusBaseWidget.__init__(self, "TaurusTrend2DDialog")
        # support x_axis_mode values (map them to stackMode values)
        stackMode = dict(
            t='datetime', d='deltatime', e='event', n='event'
        ).get(stackMode, stackMode)
        self.trendItem = None
        self.buffersize = buffersize
        self._useArchiving = False
        self._stackMode = stackMode
        self.setStackMode(stackMode)
        self.setWindowFlags(Qt.Qt.Widget)
        # add some tools
        for toolklass in (TaurusModelChooserTool, AutoScrollTool,
                          AutoScaleXTool, AutoScaleYTool, AutoScaleZTool
                          ):
            self.add_tool(toolklass)
        self.get_tool(TaurusModelChooserTool).singleModel = True

        if 'x' in autoscale.lower():
            self.get_tool(AutoScaleXTool).setChecked(True)
        if 'y' in autoscale.lower():
            self.get_tool(AutoScaleYTool).setChecked(True)
        if 'z' in autoscale.lower():
            self.get_tool(AutoScaleZTool).setChecked(True)

        self.setModifiableByUser(self._modifiableByUser)
        self.setContextMenuPolicy(Qt.Qt.CustomContextMenu)

        # Config properties
        self.setModelInConfig(True)
        self.registerConfigDelegate(self.trendItem or BaseConfigurableClass(),
                                    name='trendItem')
        self.registerConfigDelegate(self.get_tool(AutoScrollTool),
                                    name='AutoScrollTool')
        self.registerConfigDelegate(self.get_tool(AutoScaleXTool),
                                    name='AutoScaleXTool')
        self.registerConfigDelegate(self.get_tool(AutoScaleYTool),
                                    name='AutoScaleYTool')
        self.registerConfigDelegate(self.get_tool(AutoScaleZTool),
                                    name='AutoScaleZTool')
        self.registerConfigProperty(self.getStackMode,
                                    self.setStackMode,
                                    'stackMode'
                                    )
        self.registerConfigProperty(self._get_axes_conf,
                                    self._set_axes_conf,
                                    'axes_confs'
                                    )
        
    def _get_axes_conf(self):
        p = self.get_plot()
        conf = {}
        for a in p.AXIS_IDS:
            c = conf[a] = {}
            c['limits'] = p.get_axis_limits(a)
            c['title'] = p.get_axis_title(a)
            c['color'] = p.get_axis_color(a)
            c['scale'] = p.get_axis_scale(a)
            c['unit'] = p.get_axis_unit(a)
        return conf

    def _set_axes_conf(self, axes_conf):
        p = self.get_plot()
        for a in p.AXIS_IDS:
            c = axes_conf[a]
            if 'limits' in c:
                p.set_axis_limits(a, *c['limits'])
            if 'title' in c:
                p.set_axis_title(a, c['title'])
            if 'color' in c:
                p.set_axis_color(a, c['color'])
            if 'scale' in c:
                p.set_axis_scale(a, c['scale'], autoscale=False)
            if 'unit' in c:
                p.set_axis_unit(a, c['unit'])

    def keyPressEvent(self, event):
        if(event.key() == Qt.Qt.Key_Escape):
            event.ignore()
        else:
            ImageDialog.keyPressEvent(self, event)

    def setStackMode(self, mode):
        """set the type of stack to be used. This determines how X values are
        interpreted:

            - as timestamps ('datetime' or 't')
            - as time deltas ('deltatime' or 'd')
            - as event numbers ('event' or 'e')

        :param mode:(one of 'datetime', 'timedelta' or 'event')
        """
        mode = str(mode)
        if mode == 'datetime':
            self.add_tool(TimeAxisTool)
            timetool = self.get_tool(TimeAxisTool)
            timetool.set_scale_y_t(True)
        elif mode == 'deltatime':
            from taurus.qt.qtgui.extra_guiqwt.scales import DeltaTimeScaleEngine
            plot = self.get_plot()
            DeltaTimeScaleEngine.enableInAxis(plot, plot.xBottom, rotation=-45)
        elif mode == 'event':
            plot = self.get_plot()
            scaleEngine = plot.axisScaleEngine(plot.xBottom)
            if hasattr(scaleEngine, 'disableInAxis'):
                scaleEngine.disableInAxis(plot, plot.xBottom)
        else:
            self.error('Unknown stack mode "%s"' % repr(mode))
            return
        self._stackMode = mode
        if hasattr(self.trendItem, 'stackMode'):
            self.trendItem.stackMode = mode

    def getStackMode(self):
        return self._stackMode

    def resetStackMode(self):
        self.setStackMode('datetime')

    def getModelClass(self):
        """reimplemented from :class:`TaurusBaseWidget`"""
        return taurus.core.taurusattribute.TaurusAttribute

    def setModel(self, model):
        """reimplemented from :class:`TaurusBaseWidget`"""
        plot = self.get_plot()
        if self.trendItem is not None:
            plot.del_item(self.trendItem)
        self.trendItem = TaurusTrend2DItem(
            stackMode=self.getStackMode(), buffersize=self.buffersize)
        self.trendItem.setModel(model)
        plot.add_item(self.trendItem, autoscale=False)
        self.trendItem.set_readonly(not self.isModifiableByUser())
        plot.set_axis_title(plot.colormap_axis, 'value')
        plot.set_axis_unit('left', 'index')
        try:
            plot.set_axis_title(
                'left', self.trendItem.getModelObj().getSimpleName())
        except:
            self.debug('cannot set title for left axis')
            self.traceback()
        try:
            unit = self.trendItem.getModelObj().getConfig().getUnit() or ''
            plot.set_axis_unit(plot.colormap_axis, unit)
        except:
            self.debug('cannot set units for colormap axis')
            self.traceback()

        self.trendItem.dataChanged.connect(self.update_cross_sections)

        # unregister old trendItem and register the new one as config delegate
        self.unregisterConfigurableItem ("trendItem", raiseOnError=False)
        if self.getModelInConfig():
            self.registerConfigDelegate(self.trendItem, name='trendItem')

    def getModel(self):
        """reimplemented from :class:`TaurusBaseWidget`"""
        if self.trendItem is None:
            return None
        else:
            return self.trendItem.getModel()

    def setUseArchiving(self, enable):
        """enables/disables looking up in the archiver for data stored before
        the Trend was started

        :param enable: (bool) if True, archiving values will be used if
                       available
        """
        if not self._stackMode == 'datetime':
            self.info('ignoring setUseArchiving. Reason: not in X time scale')
        self._useArchiving = enable

    def getUseArchiving(self):
        """whether TaurusTrend is looking for data in the archiver when needed

        :return: (bool)

        .. seealso:: :meth:`setUseArchiving`
        """
        return self._useArchiving

    def resetUseArchiving(self):
        """Same as setUseArchiving(False)"""
        self.setUseArchiving(False)

    def setMaxDataBufferSize(self, maxSize):
        """sets the maximum number of events that will be stacked

        :param maxSize: (int) the maximum limit

        .. seealso:: :class:`TaurusTrendSet`
        """
        if self.trendItem is not None:
            self.trendItem.setBufferSize(maxSize)

        self.buffersize = maxSize

    def getMaxDataBufferSize(self):
        """returns the maximum number of events that can be plotted in the trend

        :return: (int)
        """
        return self.buffersize

    def resetMaxDataBufferSize(self):
        """Same as setMaxDataBufferSize(512)  (i.e. 512 events)"""
        self.setMaxDataBufferSize(512)

    @classmethod
    def getQtDesignerPluginInfo(cls):
        """reimplemented from :class:`TaurusBaseWidget`"""
        ret = TaurusBaseWidget.getQtDesignerPluginInfo()
        ret['module'] = 'taurus.qt.qtgui.extra_guiqwt'
        ret['group'] = 'Taurus Display'
        ret['icon'] = 'designer:qwtplot.png'
        return ret

    def setModifiableByUser(self, modifiable):
        """reimplemented from :class:`TaurusBaseWidget`"""
        self.get_tool(TaurusModelChooserTool).action.setEnabled(modifiable)
        self.get_plot().set_items_readonly(not modifiable)
        TaurusBaseWidget.setModifiableByUser(self, modifiable)

    model = Qt.pyqtProperty("QString", getModel, setModel,
                            TaurusBaseWidget.resetModel)
    # @todo uncomment this when archiving is supported
    useArchiving = Qt.pyqtProperty("bool", getUseArchiving, setUseArchiving,
                                   resetUseArchiving)
    maxDataBufferSize = Qt.pyqtProperty("int", getMaxDataBufferSize,
                                        setMaxDataBufferSize,
                                        resetMaxDataBufferSize)
    stackMode = Qt.pyqtProperty("QString", getStackMode, setStackMode,
                                resetStackMode)
    modifiableByUser = Qt.pyqtProperty("bool",
                                       TaurusBaseWidget.isModifiableByUser,
                                       setModifiableByUser,
                                       TaurusBaseWidget.resetModifiableByUser)



# -*- coding: utf-8 -*-

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""Unit tests for UILoadable decorator"""

from __future__ import absolute_import

import os.path

import unittest
from taurus.external.qt import Qt
from taurus.qt.qtgui.util.ui import UILoadable
from .mywidget3 import MyWidget3


class UILoadableTestCase(unittest.TestCase):
    """
    Test cases for UILoadable decorator
    """

@UILoadable
class MyWidget1(Qt.QWidget):

    def __init__(self, parent=None):
        Qt.QWidget.__init__(self, parent)
        self.loadUi()
        self.my_button.setText("This is MY1 button")


@UILoadable(with_ui="ui")
class MyWidget2(Qt.QWidget):

    def __init__(self, parent=None):
        Qt.QWidget.__init__(self, parent)
        path = os.path.join(os.path.dirname(__file__), "ui", "mywidget2")
        self.loadUi(filename="mywidget2_custom.ui", path=path)
        self.ui.my_button.setText("This is MY2 button")


class Bug151_Widget(MyWidget3):
    pass


def test_uiloadable_default(qtbot):
    """Test UILoadable with default arguments"""
    w = MyWidget1()
    qtbot.addWidget(w)
    assert w.my_button.text() == "This is MY1 button"


def test_uiloadable_customized(qtbot):
    """Test UILoadable with customized filename and path"""
    w = MyWidget2()
    qtbot.addWidget(w)
    assert hasattr(w, "ui")
    assert hasattr(w.ui, "my_button")
    assert not hasattr(w, "my_button")
    assert w.ui.my_button.text() == "This is MY2 button"


def test_bug151(qtbot):
    """Check inheritance of UILoadable classes across packages (bug #151)"""

    w = Bug151_Widget()
    qtbot.addWidget(w)



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""This module provides a panel to display taurus messages"""

__all__ = ["ClipboardReportHandler", "SMTPReportHandler"]

__docformat__ = 'restructuredtext'

from taurus.core.util.report import TaurusMessageReportHandler
from taurus.external.qt import Qt
from taurus.qt.qtgui.util.ui import UILoadable


class ClipboardReportHandler(TaurusMessageReportHandler):
    """Report a message by copying it to the clipboard"""

    Label = "Copy to Clipboard"

    def report(self, message):
        app = Qt.QApplication.instance()
        clipboard = app.clipboard()
        clipboard.setText(message)

        Qt.QMessageBox.information(None, "Done!",
                                   "Message Copied to clipboard")


@UILoadable(with_ui='ui')
class SendMailDialog(Qt.QDialog):

    def __init__(self, parent=None):
        Qt.QDialog.__init__(self, parent)
        self.loadUi(filename="SendMailForm.ui")
        self.ui.buttonBox.accepted.connect(self.accept)
        self.ui.buttonBox.rejected.connect(self.reject)
        self.ui.editMessage.setFont(Qt.QFont("Monospace"))

    def setFrom(self, efrom):
        self.ui.lineEditFrom.setText(efrom)

    def setTo(self, eto):
        self.ui.editTo.setText(eto)

    def setSubject(self, subject):
        self.ui.editSubject.setText(subject)

    def setMessage(self, message):
        self.ui.editMessage.setPlainText(message)

    def getFrom(self):
        return str(self.ui.editFrom.text())

    def getTo(self):
        return str(self.ui.editTo.text())

    def getSubject(self):
        return str(self.ui.editSubject.text())

    def getMessage(self):
        return str(self.ui.editMessage.toPlainText())

    def getMailInfo(self):
        return self.getFrom(), self.getTo(), self.getSubject(), \
            self.getMessage()


class SMTPReportHandler(TaurusMessageReportHandler):
    """Report a message by sending an email"""

    Label = "Send email"

    def report(self, message):

        app = Qt.QApplication.instance()

        subject = "Error in " + app.applicationName()
        dialog = self.createDialog(subject=subject, message=message)

        if not dialog.exec_():
            return

        mail_info = dialog.getMailInfo()

        try:
            self.sendMail(*mail_info)
            Qt.QMessageBox.information(None, "Done!",
                                       "Email has been sent!")
        except:
            import sys
            import traceback
            einfo = sys.exc_info()[:2]
            msg = "".join(traceback.format_exception_only(*einfo))
            Qt.QMessageBox.warning(None, "Failed to send email",
                                   "Failed to send email. Reason:\n\n" + msg)

    def sendMail(self, efrom, eto, subject, message):
        import smtplib
        import email.mime.text
        msg = email.mime.text.MIMEText(message)
        msg['From'] = efrom
        msg['To'] = eto
        msg['Subject'] = subject

        s = smtplib.SMTP('localhost')
        s.sendmail(efrom, eto, msg.as_string())
        s.quit()

    def getDialogClass(self):
        return SendMailDialog

    def createDialog(self, efrom=None, eto=None, subject=None, message=None):
        dialog = self.getDialogClass()()
        dialog.setWindowTitle("Compose message")
        if efrom is not None:
            dialog.setFrom(efrom)
        if eto is not None:
            dialog.setFrom(eto)
        if subject is not None:
            dialog.setSubject(subject)
        if message is not None:
            dialog.setMessage(message)
        return dialog


def main():
    app = Qt.QApplication([])
    w = SendMailDialog()
    w.exec_()

if __name__ == "__main__":
    main()



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""
This module contains a set of useful containers that are not part of the standard
python distribution.
"""
from __future__ import print_function

from builtins import range
from builtins import object

from future.utils import string_types

import copy
import time
import weakref
try:
    from collections.abc import Sequence
except ImportError:  # bck-compat py 2.7
    from collections import Sequence

__all__ = ["CaselessList", "CaselessDict", "CaselessWeakValueDict", "LoopList",
           "CircBuf", "LIFO", "TimedQueue", "self_locked", "ThreadDict",
           "defaultdict", "defaultdict_fromkey", "CaselessDefaultDict",
           "DefaultThreadDict", "getDictAsTree", "ArrayBuffer", ]

__docformat__ = "restructuredtext"


class CaselessList(list):
    """A case insensitive lists that has some caseless methods. Only allows
    strings as list members. Most methods that would normally return a list,
    return a CaselessList. (Except list() and lowercopy())
    Sequence Methods implemented are :
    __contains__, remove, count, index, append, extend, insert,
    __getitem__, __setitem__, __getslice__, __setslice__
    __add__, __radd__, __iadd__, __mul__, __rmul__
    Plus Extra methods:
    findentry, copy , lowercopy, list
    Inherited methods :
    __imul__, __len__, __iter__, pop, reverse, sort
    """

    def __init__(self, inlist=[]):
        list.__init__(self)
        for entry in inlist:
            if not isinstance(entry, string_types):
                raise TypeError('Members of this object must be strings. '
                                'You supplied \"%s\" which is \"%s\"' %
                                (entry, type(entry)))
            self.append(entry)

    def __lowerstreq(self, a, b):
        a = str(a)
        b = str(b)
        return (a.lower() == b.lower())

    def findentry(self, item):
        """A caseless way of checking if an item is in the list or not.
        It returns None or the entry."""
        if not isinstance(item, string_types):
            raise TypeError('Members of this object must be strings. '
                            'You supplied \"%s\"' % type(item))
        for entry in self:
            if self.__lowerstreq(item, entry):
                return entry
        return None

    def __contains__(self, item):
        """A caseless way of checking if a list has a member in it or not."""
        for entry in self:
            if self.__lowerstreq(item, entry):
                return True
        return False

    def remove(self, item):
        """Remove the first occurence of an item, the caseless way."""
        for entry in self:
            if self.__lowerstreq(item, entry):
                list.remove(self, entry)
                return
        raise ValueError(': list.remove(x): x not in list')

    def copy(self):
        """Return a CaselessList copy of self."""
        return CaselessList(self)

    def list(self):
        """Return a normal list version of self."""
        return list(self)

    def lowercopy(self):
        """Return a lowercase (list) copy of self."""
        return [str(entry).lower() for entry in self]

    def append(self, item):
        """Adds an item to the list and checks it's a string."""
        if not isinstance(item, string_types):
            raise TypeError('Members of this object must be strings. '
                            'You supplied \"%s\"' % type(item))
        list.append(self, item)

    def extend(self, item):
        """Extend the list with another list. Each member of the list must be
        a string."""
        if not isinstance(item, list):
            raise TypeError('You can only extend lists with lists. '
                            'You supplied \"%s\"' % type(item))
        for entry in item:
            if not isinstance(entry, string_types):
                raise TypeError('Members of this object must be strings. '
                                'You supplied \"%s\"' % type(entry))
            list.append(self, entry)

    def count(self, item):
        """Counts references to 'item' in a caseless manner.
        If item is not a string it will always return 0."""
        if not isinstance(item, string_types):
            return 0
        count = 0
        for entry in self:
            if self.__lowerstreq(item, entry):
                count += 1
        return count

    def index(self, item, minindex=0, maxindex=None):
        """Provide an index of first occurence of item in the list. (or raise
        a ValueError if item not present)
        If item is not a string, will raise a TypeError.
        minindex and maxindex are also optional arguments
        s.index(x[, i[, j]]) return smallest k such that s[k] == x and i <= k < j
        """
        if maxindex is None:
            maxindex = len(self)
        minindex = max(0, minindex) - 1
        maxindex = min(len(self), maxindex)
        if not isinstance(item, string_types):
            raise TypeError('Members of this object must be strings. '
                            'You supplied \"%s\"' % type(item))
        index = minindex
        while index < maxindex:
            index += 1
            if self.__lowerstreq(item, self[index]):
                return index
        raise ValueError(': list.index(x): x not in list')

    def insert(self, i, x):
        """s.insert(i, x) same as s[i:i] = [x]
        Raises TypeError if x isn't a string."""
        if not isinstance(x, string_types):
            raise TypeError('Members of this object must be strings. '
                            'You supplied \"%s\"' % type(x))
        list.insert(self, i, x)

    def __setitem__(self, index, value):
        """For setting values in the list.
        index must be an integer or (extended) slice object. (__setslice__ used
        for simple slices)
        If index is an integer then value must be a string.
        If index is a slice object then value must be a list of strings - with
        the same length as the slice object requires.
        """
        if isinstance(index, int):
            if not isinstance(value, string_types):
                raise TypeError('Members of this object must be strings. '
                                'You supplied \"%s\"' % type(value))
            list.__setitem__(self, index, value)
        elif isinstance(index, slice):
            if not hasattr(value, '__len__'):
                raise TypeError(
                    'Value given to set slice is not a sequence object.')
            for entry in value:
                if not isinstance(entry, string_types):
                    raise TypeError('Members of this object must be strings. '
                                    'You supplied \"%s\"' % type(entry))
            list.__setitem__(self, index, value)
        else:
            raise TypeError('Indexes must be integers or slice objects.')

    def __setslice__(self, i, j, sequence):
        """Called to implement assignment to self[i:j]."""
        for entry in sequence:
            if not isinstance(entry, string_types):
                raise TypeError('Members of this object must be strings. '
                                'You supplied \"%s\"' % type(entry))
        list.__setslice__(self, i, j, sequence)

    def __getslice__(self, i, j):
        """Called to implement evaluation of self[i:j].
        Although the manual says this method is deprecated - if I don't define
        it the list one is called.
        (Which returns a list - this returns a CaselessList)"""
        return CaselessList(list.__getslice__(self, i, j))

    def __getitem__(self, index):
        """For fetching indexes.
        If a slice is fetched then the list returned is a CaselessList."""
        if not isinstance(index, slice):
            return list.__getitem__(self, index)
        else:
            return CaselessList(list.__getitem__(self, index))

    def __add__(self, item):
        """To add a list, and return a CaselessList.
        Every element of item must be a string."""
        return CaselessList(list.__add__(self, item))

    def __radd__(self, item):
        """To add a list, and return a CaselessList.
        Every element of item must be a string."""
        return CaselessList(list.__add__(self, item))

    def __iadd__(self, item):
        """To add a list in place."""
        for entry in item:
            self.append(entry)

    def __mul__(self, item):
        """To multiply itself, and return a CaselessList.
        Every element of item must be a string."""
        return CaselessList(list.__mul__(self, item))

    def __rmul__(self, item):
        """To multiply itself, and return a CaselessList.
        Every element of item must be a string."""
        return CaselessList(list.__rmul__(self, item))


class CaselessDict(dict):
    """A case insensitive dictionary. Use this class as a normal dictionary.
    The keys must be strings"""

    def __init__(self, other=None):
        if other:
            # Doesn't do keyword args

            # -------------------------------------------------------
            # TODO: when we drop py2 support, change this to
            #       `if isinstance(other, collections.abc.Mapping)`
            #
            if hasattr(other, 'items'):
                for k, v in other.items():
                    dict.__setitem__(self, k.lower(), v)
            # -------------------------------------------------------
            else:
                for k, v in other:
                    dict.__setitem__(self, k.lower(), v)

    def __getitem__(self, key):
        return dict.__getitem__(self, key.lower())

    def __setitem__(self, key, value):
        dict.__setitem__(self, key.lower(), value)

    def __contains__(self, key):
        return dict.__contains__(self, key.lower())

    def has_key(self, key):
        """overwritten from :meth:`dict.has_key` (needed for python2)"""
        return key.lower() in self

    def get(self, key, def_val=None):
        """overwritten from :meth:`dict.get`"""
        return dict.get(self, key.lower(), def_val)

    def setdefault(self, key, def_val=None):
        """overwritten from :meth:`dict.setdefault`"""
        return dict.setdefault(self, key.lower(), def_val)

    def update(self, other):
        """overwritten from :meth:`dict.update`"""
        for k, v in other.items():
            dict.__setitem__(self, k.lower(), v)

    def fromkeys(self, iterable, value=None):
        d = CaselessDict()
        for k in iterable:
            dict.__setitem__(d, k.lower(), value)
        return d

    def pop(self, key, def_val=None):
        """overwritten from :meth:`dict.pop`"""
        return dict.pop(self, key.lower(), def_val)

    def __delitem__(self, k):
        dict.__delitem__(self, k.lower())


class CaselessWeakValueDict(weakref.WeakValueDictionary):

    def __init__(self, other=None):
        weakref.WeakValueDictionary.__init__(self, other)
        if other:
            # Doesn't do keyword args
            if isinstance(other, dict):
                for k, v in other.items():
                    weakref.WeakValueDictionary.__setitem__(self, k.lower(), v)
            else:
                for k, v in other:
                    weakref.WeakValueDictionary.__setitem__(self, k.lower(), v)

    def __getitem__(self, key):
        return weakref.WeakValueDictionary.__getitem__(self, key.lower())

    def __setitem__(self, key, value):
        weakref.WeakValueDictionary.__setitem__(self, key.lower(), value)

    def __contains__(self, key):
        return weakref.WeakValueDictionary.__contains__(self, key.lower())

    def has_key(self, key):
        """overwritten from :meth:`weakref.WeakValueDictionary`
        (needed for python2)
        """
        return key in self

    def get(self, key, def_val=None):
        """overwritten from :meth:`weakref.WeakValueDictionary.get`"""
        return weakref.WeakValueDictionary.get(self, key.lower(), def_val)

    def setdefault(self, key, def_val=None):
        """overwritten from :meth:`weakref.WeakValueDictionary.setdefault`"""
        return weakref.WeakValueDictionary.setdefault(self, key.lower(), def_val)

    def update(self, other):
        """overwritten from :meth:`weakref.WeakValueDictionary.update`"""
        if other:
            for k, v in other.items():
                weakref.WeakValueDictionary.__setitem__(self, k.lower(), v)

    def fromkeys(self, iterable, value=None):
        d = CaselessWeakValueDict()
        for k in iterable:
            weakref.WeakValueDictionary.__setitem__(d, k.lower(), value)
        return d

    def pop(self, key, def_val=None):
        """overwritten from :meth:`weakref.WeakValueDictionary.pop`"""
        return weakref.WeakValueDictionary.pop(self, key.lower(), def_val)

    def __delitem__(self, k):
        weakref.WeakValueDictionary.__delitem__(self, k.lower())


# {{{ http://code.activestate.com/recipes/576642/ (r10)
import pickle
import json
import csv
import os
import shutil


class PersistentDict(dict):
    ''' Persistent dictionary with an API compatible with shelve and anydbm.

    The dict is kept in memory, so the dictionary operations run as fast as
    a regular dictionary.

    Write to disk is delayed until close or sync (similar to gdbm's fast mode).

    Input file format is automatically discovered.
    Output file format is selectable between pickle, json, and csv.
    All three serialization formats are backed by fast C implementations.

    '''

    def __init__(self, filename, flag='c', mode=None, format='pickle', *args, **kwds):
        self.flag = flag                    # r=readonly, c=create, or n=new
        self.mode = mode                    # None or an octal triple like 0644
        self.format = format                # 'csv', 'json', or 'pickle'
        self.filename = filename
        if flag != 'n' and os.access(filename, os.R_OK):
            fileobj = open(filename, 'rb' if format == 'pickle' else 'r')
            with fileobj:
                self.load(fileobj)
        dict.__init__(self, *args, **kwds)

    def sync(self):
        'Write dict to disk'
        if self.flag == 'r':
            return
        filename = self.filename
        tempname = filename + '.tmp'
        fileobj = open(tempname, 'wb' if self.format == 'pickle' else 'w')
        try:
            self.dump(fileobj)
        except Exception:
            os.remove(tempname)
            raise
        finally:
            fileobj.close()
        shutil.move(tempname, self.filename)    # atomic commit
        if self.mode is not None:
            os.chmod(self.filename, self.mode)

    def close(self):
        self.sync()

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        self.close()

    def dump(self, fileobj):
        if self.format == 'csv':
            csv.writer(fileobj).writerows(list(self.items()))
        elif self.format == 'json':
            json.dump(self, fileobj, separators=(',', ':'))
        elif self.format == 'pickle':
            pickle.dump(dict(self), fileobj, 2)
        else:
            raise NotImplementedError('Unknown format: ' + repr(self.format))

    def load(self, fileobj):
        # try formats from most restrictive to least restrictive
        for loader in (pickle.load, json.load, csv.reader):
            fileobj.seek(0)
            try:
                return self.update(loader(fileobj))
            except Exception:
                pass
        raise ValueError('File not in a supported format')


class LoopList(object):
    '''this class provides an effectively cyclic list.
    It can be used, e.g., for storing colors or pen properties to be changed automatically in a plot

    A LoopList stores an internal index to remember the last accessed item in the list
    It provides  previous(), current() and next() methods that return the previous,current and next items in the list.
    The method allItems() returns a copy of all items contained in the list.
    The index can be accessed by setCurrentIndex() and getCurrentIndex()  (setCurrentIndex(i) additionally returns new current item)
    Items can be accessed ***without modifying the current index*** by using llist[i]  and llist[i]=x syntax
    len(llist) returns the **period** of the list.

    .. note::

        only basic methods of lists are implemented for llists. In particular,
        the following are **not** implemented:

            - slicing
            - resizing (append, insert, del,...)
            - binary operators (+,*,...)

    ..note::

        it can be used for loops, but the loop will be infinite unless other
        condition is used for exiting it:

            - for item in llist: print item  # This is a infinite loop!!
            - for i in range(len(llist)):print  llist[i]  #This is not infinite
              since len(llist) returns the period of the list'''

    def __init__(self, itemlist=[]):
        self.setItemList(itemlist)

    def setItemList(self, itemlist):
        '''sets the item list'''
        self._itemlist = list(itemlist)
        self._index = 0
        self._nitems = len(self._itemlist)

    def current(self):
        '''returns current item'''
        try:
            # makes the list effectively cyclic
            return self._itemlist[self._index % self._nitems]
        except ZeroDivisionError:
            raise IndexError('cyclic list is empty')

    def setCurrentIndex(self, index):
        '''sets current index (and returns the corresponding item)'''
        self._index = index
        return self.current()

    def getCurrentIndex(self):
        '''returns the current index'''
        return self._index

    def __next__(self):
        '''advances one item in the list and returns it'''
        self._index += 1
        return self.current()

    next = __next__

    def previous(self):
        '''goes one item back in the list and returns it'''
        self._index -= 1
        return self.current()

    def allItems(self):
        """returns the items list (one period)"""
        return self._itemlist

    def __getitem__(self, i):
        try:
            return self._itemlist[i % self._nitems]
        except ZeroDivisionError:
            raise IndexError('cyclic list is empty')

    def __setitem__(self, i, y):
        self._itemlist[i % self._nitems] = y

    def __len__(self):
        return self._nitems


class CircBuf(object):
    """A circular buffer of Python values.

    Examples::

        >>> cb = CircBuf(3)
        >>> cb.is_empty()
        1
        >>> cb.put('first')
        >>> cb.is_empty()
        0
        >>> cb.put('second')
        >>> cb.put('third')
        >>> cb.is_full()
        1
        >>> cb.put('fourth')
        >>> cb.get()
        'second'
        >>> cb.get()
        'third'
        >>> cb.get()
        'fourth'
        >>> cb.is_empty()
        1"""

    def __init__(self, leng):
        """Construct an empty circular buffer."""

        self.buf = leng * [None]
        self.len = self.g = self.p = 0

    def is_empty(self):
        """Returns true only if CircBuf has no items."""
        return self.len == 0

    def is_full(self):
        """Returns true only if CircBuf has no space."""
        return self.len == len(self.buf)

    def get(self):
        """Retrieves an item from a non-empty circular buffer."""
        result = self.buf[self.g]
        self.g = (self.g + 1) % len(self.buf)
        self.len -= 1
        return result

    def put(self, item):
        """Puts an item onto a circular buffer."""
        self.buf[self.p] = item
        self.p = (self.p + 1) % len(self.buf)
        if self.len == len(self.buf):
            self.g = self.p
        else:
            self.len += 1


class LIFO(object):

    def __init__(self, max=10):
        self._data = []
        self._max = max

    def append(self, elem):
        if self._max == 0:
            return

        while len(self._data) >= self._max:
            self._data.pop(0)
        self._data.append(elem)

    def extend(self, lst):
        if self._max == 0:
            return
        s = len(lst)
        if s >= self._max:
            self._data = lst[s - self._max:]
        else:
            while len(self._data) + len(lst) > self._max:
                self._data.pop(0)
            self._data.extend(lst)

    def pop(self, index=0):
        self._data.pop(index)

    def clear(self):
        self._data = []

    def get(self):
        return self._data

    def getCopy(self):
        return copy.copy(self._data)

    def __str__(self):
        return "%s(%s)" % (self.__class__.__name__, str(self._data))


class TimedQueue(list):
    """ A FIFO that keeps all the values introduced at least for a given time.
    Applied to some device servers, to force States to be kept at least a minimum time.
    Previously named as PyTango_utils.device.StateQueue
    pop(): The value is removed only if delete_time has been reached.
    at least 1 value is always kept in the list
    """

    def __init__(self, arg=None):
        """ Initializes the list with a sequence or an initial value. """
        if arg is None:
            list.__init__(self)
        elif isinstance(arg, Sequence):
            list.__init__(self, arg)
        else:
            list.__init__(self)
            self.append(arg, 1)

    def append(self, obj, keep=15):
        """ Inserts a tuple with (value,insert_time,delete_time=now+keep) """
        now = time.time()
        l = (obj, now, now + keep)
        list.append(self, l)

    def pop(self, index=0):
        """ Returns the indicated value, or the first one; but removes only if delete_time has been reached.
        All values are returned at least once.
        When the queue has only a value, it is not deleted.
        """
        if not self:
            return None  # The list is empty
        now = time.time()
        s, t1, t2 = self[index]
        if now < t2 or len(self) == 1:
            return s
        else:
            return list.pop(self, index)[0]

    def index(self, obj):
        for i in range(len(self)):
            if self[i][0] == obj:
                return i
        return None

    def __contains__(self, obj):
        for t in self:
            if t[0] == obj:
                return True
        return False
    pass


def self_locked(func, reentrant=True):
    '''Decorator to make thread-safe class members
       Decorator to create thread-safe objects.

       .. warning::

            - With Lock() this decorator should not be used to decorate
              nested functions; it will cause Deadlock!
            - With RLock this problem is avoided ... but you should rely more
              on python threading'''
    import threading

    def lock_fun(self, *args, **kwargs):
        if not hasattr(self, 'lock'):
            if reentrant:
                setattr(self, 'lock', threading.RLock())
            else:
                setattr(self, 'lock', threading.Lock())
        if not hasattr(self, 'trace'):
            setattr(self, 'trace', False)
        self.lock.acquire()
        try:
            if self.trace:
                print("locked: %s" % self.lock)
            result = func(self, *args, **kwargs)
        finally:
            self.lock.release()
            if self.trace:
                print("released: %s" % self.lock)
        return result
    return lock_fun


class ThreadDict(dict):
    ''' Thread safe dictionary with redefinable read/write methods and a
    background thread for hardware update.  All methods are thread-safe using
    *@self_lock* decorator.

    .. note::
        any method decorated in this way CANNOT call other decorated methods!
        All values of the dictionary will be automatically updated in a separate
        Thread using read_method provided. Any value overwritten in the dict
        should launch the write_method.

    Briefing::

        a[2] equals to a[2]=read_method(2)
        a[2]=1 equals to a[2]=write_method(2,1)'''

    def __init__(self, other=None, read_method=None, write_method=None, timewait=0.1, threaded=True):
        self.read_method = read_method
        self.write_method = write_method
        self.timewait = timewait
        self.threaded = threaded
        self._threadkeys = []
        self.trace = False
        self.last_update = 0
        self.last_cycle_start = 0
        # equals to self.__class__.__base__ or type(self).__bases__[0]
        self.parent = type(self).mro()[1]

    def tracer(self, text):
        print(text)

    def start(self):
        import threading
        if not self.threaded:
            return
        if hasattr(self, '_Thread') and self._Thread and self._Thread.is_alive():
            return
        self.event = threading.Event()
        self.event.clear()
        self._Thread = threading.Thread(target=self.run)
        self._Thread.setDaemon(True)
        self._Thread.start()

    def stop(self):
        if self.threaded and hasattr(self, 'event'):
            self.event.set()
        if hasattr(self, '_Thread'):
            self._Thread.join()

    def alive(self):
        if not hasattr(self, '_Thread') or not self._Thread:
            return False
        else:
            return self._Thread.is_alive()

    def __del__(self):
        self.stop()
        # dict.__del__(self)

    def run(self):
        while not self.event.isSet():
            keys = self.threadkeys()
            for k in keys:
                try:
                    #@todo It must be checked wich method for reading is better for thread_safe
                    if True:
                        self.__getitem__(k, hw=True)
                    # try this alternative in case of deadlock (it could need
                    # extra locks inside read_method)
                    else:
                        if self.read_method:
                            value = self.read_method(k)
                            self.__setitem__(k, value, hw=False)
                finally:
                    self.event.wait(self.timewait)
                if self.event.isSet():
                    break
            timewait = self.get_timewait()
            self.event.wait(timewait)

    @self_locked
    def get_last_update(self):
        return self.last_update

    @self_locked
    def set_last_update(self, value):
        self.last_update = value

    @self_locked
    def get_last_cycle_start(self):
        return self.last_cycle_start

    @self_locked
    def set_last_cycle_start(self, value):
        self.last_cycle_start = value

    @self_locked
    def get_timewait(self):
        return self.timewait

    @self_locked
    def set_timewait(self, value):
        self.timewait = value

    @self_locked
    def append(self, key, value=None):
        if key not in self:
            self.parent.__setitem__(self, key, value)
        if key not in self._threadkeys:
            self._threadkeys.append(key)

    @self_locked
    def threadkeys(self):
        return self._threadkeys[:]

    @self_locked
    def __getitem__(self, key, hw=False):
        ''' This method launches a read_method execution if there's no thread on charge of doing that or if the hw flag is set to True. '''
        import time
        if (hw or not self.threaded) and self.read_method:
            dict.__setitem__(self, key, self.read_method(key))
            self.last_update = time.time()
        return dict.__getitem__(self, key)

    @self_locked
    def __setitem__(self, key, value, hw=True):
        ''' This method launches a write_method execution if the hw flag is not explicitly set to False. '''
        import time
        if hw and self.write_method:
            # It implies that a key will not be added here to read thread!
            dict.__setitem__(self, key, self.write_method(*[key, value]))
        else:
            dict.__setitem__(self, key, value)
        self.last_update = time.time()

    @self_locked
    def get(self, key, default=None):
        import time
        if not self.threaded and self.read_method:
            dict.__setitem__(self, key, self.read_method(key))
            self.last_update = time.time()
        if default is False:
            return dict.get(self, key)
        else:
            return dict.get(self, key, default)

    #__getitem__ = self_locked(dict.__getitem__)
    #__setitem__ = self_locked(dict.__setitem__)
    __delitem__ = self_locked(dict.__delitem__)
    __contains__ = self_locked(dict.__contains__)
    __iter__ = self_locked(dict.__iter__)

    pop = self_locked(dict.pop)
    #@self_locked
    # def pop(self,k,d=None):
    # if k in self.keys():
    # self.stop()
    #d = dict.pop(self,k)
    # self.start()
    # return d

    @self_locked
    def __str__(self):
        return "{" + ",".join(["'" + str(k) + "'" + ":" + "'" + str(v) + "'" for k, v in zip(dict.keys(self), dict.values(self))]) + "}"

    @self_locked
    def __repr__(self):
        return "{\n" + "\n,".join(["'" + str(k) + "'" + ":" + "'" + str(v) + "'" for k, v in zip(dict.keys(self), dict.values(self))]) + "\n}"
    #__str__ = self_locked(dict.__str__)
    #__repr__ = self_locked(dict.__repr__)

    #get = self_locked(dict.get)
    #has_key = self_locked(dict.has_key)
    update = self_locked(dict.update)
    copy = self_locked(dict.copy)

    keys = self_locked(dict.keys)
    values = self_locked(dict.values)
    items = self_locked(dict.items)
    #iterkeys = self_locked(dict.iterkeys)
    #itervalues = self_locked(dict.itervalues)
    #iteritems = self_locked(dict.iteritems)


class SortedDict(dict):
    """ This class implements a dictionary that returns keys in the same order they were inserted. """

    def __init__(self, other=None):
        dict.__init__(self)
        self._keys = []
        if other is not None:
            self.update(other)
        return

    def sort(self, key):
        """
        This method modifies the sorting of the dictionary overriding the existing sort key.
        :param key: it can be a sequence containing all the keys already existing in the dictionary
                    or a callable providing a sorting key algorithm.
        """
        import operator
        if hasattr(key, '__call__'):
            self._keys = sorted(self._keys, key=key)
        else:
            for k in self._keys:
                if k not in self._keys:
                    raise KeyError(k)
            self._keys = list(key)
        return self._keys[:]

    def __setitem__(self, k, v):
        if k not in self._keys:
            self._keys.append(k)
        dict.__setitem__(self, k, v)

    def update(self, other):
        if hasattr(other, 'items'):
            other = other.items()
        for k, v in other:
            self.__setitem__(k, v)

    @staticmethod
    def fromkeys(S, v=None):
        return SortedDict((s, v) for s in S)

    def pop(self, k, d=None):
        """Removes key and returns its (self[key] or d or None)"""
        if k not in self:
            return d
        self._keys.remove(k)
        return dict.pop(self, k)

    def popitem(self):
        """Removes and returns last key,value pair"""
        k = self._keys[-1]
        v = self[k]
        self.pop(k)
        return k, v

    def clear(self):
        self._keys = []
        return dict.clear(self)

    def keys(self):
        return self._keys[:]

    def values(self):
        return [self[k] for k in self._keys]

    def items(self):
        return [(k, self[k]) for k in self._keys]

    def __iter__(self):
        return (i for i in self._keys)

    def iteritems(self):
        return ((k, self[k]) for k in self._keys)

    def iterkeys(self):
        return (i for i in self._keys)

    def itervalues(self):
        return (self[k] for k in self._keys)

try:
    from collections import defaultdict
except:
    class defaultdict(dict):

        def __init__(self, default_factory=None, *a, **kw):
            if (default_factory is not None and
                    not hasattr(default_factory, '__call__')):
                raise TypeError('first argument must be callable')
            dict.__init__(self, *a, **kw)
            self.default_factory = default_factory

        def __getitem__(self, key):
            try:
                return dict.__getitem__(self, key)
            except KeyError:
                return self.__missing__(key)

        def __missing__(self, key):
            if self.default_factory is None:
                raise KeyError(key)
            self[key] = value = self.default_factory()
            return value

        def __reduce__(self):
            if self.default_factory is None:
                args = tuple()
            else:
                args = self.default_factory,
            return type(self), args, None, None, list(self.items())

        def copy(self):
            return self.__copy__()

        def __copy__(self):
            return type(self)(self.default_factory, self)

        def __deepcopy__(self, memo):
            import copy
            return type(self)(self.default_factory,
                              copy.deepcopy(list(self.items())))

        def __repr__(self):
            return 'defaultdict(%s, %s)' % (self.default_factory,
                                            dict.__repr__(self))


class defaultdict_fromkey(defaultdict):
    """ Creates a dictionary with a default_factory function that creates new elements using key as argument.
    Usage : new_dict = defaultdict_fromkey(method); where method like (lambda key: return new_obj(key))
    Each time that new_dict[key] is called with a key that doesn't exist, method(key) is used to create the value
    Copied from PyAlarm device server
    """

    def __missing__(self, key):
        if self.default_factory is None:
            raise KeyError(key)
        self[key] = value = self.default_factory(key)
        return value


class CaselessDefaultDict(defaultdict_fromkey, CaselessDict):
    """ a join venture between caseless and default dict
    This class merges the two previous ones.
    This declaration equals to::

        CaselessDefaultDict = type('CaselessDefaultType',(CaselessDict,defaultdict_fromkey),{})
    """

    def __getitem__(self, key):
        return defaultdict_fromkey.__getitem__(self, key.lower())
    pass


class DefaultThreadDict(defaultdict_fromkey, ThreadDict):
    """a join venture between thread and default dict
    This class merges the two previous ones.
    @todo This two classes are not yet well integrated ... the way a new key is added to the dict must be rewritten explicitly.
    """

    def __init__(self, other=None, default_factory=None, read_method=None, write_method=None, timewait=0.1, threaded=True):
        defaultdict_fromkey.__init__(self, default_factory)
        ThreadDict.__init__(self, other, read_method,
                            write_method, timewait, threaded)
    pass


def getDictAsTree(dct):
    """This method will print a recursive dict in a tree-like
       shape::

           >>> print(getDictAsTree({'A':{'B':[1,2],'C':[3]}}))"""
    def add_to_level(l, d):
        lines = []
        if isinstance(d, dict):
            for k, v in d.items():
                print('with key "%s"' % k)
                lines.append([''] * l + [str(k)])
                lines += add_to_level(l + 1, v)
        elif type(d) in [list, set]:  # End of recursion
            for e in d:
                if isinstance(e, dict):
                    lines += add_to_level(l, e)
                else:
                    lines.append([''] * l + [str(e)])
        else:
            lines.append([''] * l + [str(d)])
        return lines
    ls = ['\t'.join(line) for line in add_to_level(0, dct)]
    print('lines are : \n', ls)
    return '\n'.join(ls)


class ArrayBuffer(object):
    '''A data buffer which internally uses a preallocated numpy.array.
    An ArrayBuffer will only present the actual contents, not the full internal
    buffer, so when appending or extending, it behaves as if dynamic
    reallocation was taking place.

    The contents of the class:`ArrayBuffer` can be accessed as if it was a numpy
    array (i.e slicing notation like b[2:3], b[:,2], b[-1],... are all valid).

    For retrieving the full contents, see :meth:`ArrayBuffer.contents` and
    :meth:`ArrayBuffer.toArray`

    On creation, a given initial internal buffer and a maximum size are set. If
    the maximum size is larger than the original internal buffer, this will be
    automatically grown in geometrical steps if needed to accommodate the
    contents, up to the maximum size. Once the contents fill the maximum size,
    appending or extending the contents will result in older contents being
    discarded (in a FIFO way)

    The :meth:`append` and meth:`extend` methods are designed to be cheap
    (especially if the internal buffer size is already at the maximum size), at
    the expense of memory usage'''

    def __init__(self, buffer, maxSize=0):
        '''Creator.

        :param buffer: (numpy.array) a numpy.array suitable to be used as the
                       internal buffer.

        :param maxSize: (int) Maximum size of the internal buffer. The internal
                        buffer length will be allowed to grow up to this value.
                        If maxSize=0 (default), the maximum size will be that of
                        the given buffer
        '''

        self.__buffer = buffer
        self.__end = 0
        self.__bsize = self.__buffer.shape[0]
        self.__maxSize = max(maxSize, self.__bsize)

    def __getitem__(self, i):
        return self.__buffer[:self.__end].__getitem__(i)

    def __getslice__(self, i, j):
        return self.__buffer[:self.__end].__getslice__(i, j)

    def __len__(self):
        return self.__end

    def __repr__(self):
        return "ArrayBuffer with contents = %s" % self.__buffer[:self.__end].__repr__()

    def __str__(self):
        return self.__buffer[:self.__end].__str__()

    def __bool__(self):
        return self.__buffer[:self.__end].__bool__()

    def __setitem__(self, i, x):
        self.__buffer[:self.__end].__setitem__(i, x)

    def __setslice__(self, i, j, a):
        if i >= self.__end or j > self.__end:
            raise IndexError()
        self.__buffer[:self.__end].__setslice__(i, j, a)

    def resizeBuffer(self, newlen):
        '''resizes the internal buffer'''
        if newlen < self.__end:
            self.__end = newlen
        shape = list(self.__buffer.shape)
        shape[0] = newlen
        try:
            self.__buffer.resize(shape)  # first try to resize in-place
        except:
            import numpy
            # if not possible, do it by copying
            self.__buffer = numpy.resize(self.__buffer, shape)
        self.__bsize = self.__buffer.shape[0]

    def append(self, x):
        ''' similar to the append method in a list, except that once the maximum
        buffer size is reached, elements get discarded on the begginning to keep
        the size within the limit

        :param x: (scalar) element to be appended

        .. seealso:: :meth:`extend`
        '''
        try:
            self.__buffer[self.__end] = x
        except IndexError:
            if self.__bsize < self.__maxSize:
                self.resizeBuffer(min(2 * self.__bsize, self.__maxSize))
                # recursively call itself again after resizing
                return self.append(x)
            self.moveLeft(1)
            self.__buffer[self.__end] = x
        self.__end += 1

    def extend(self, a):
        ''' similar to the extend method of a list, except that once the maximum
        buffer size is reached, elements get discarded on the begginning to keep
        the size within the limit

        :param a: (numpy.array) array of elements to append

        .. seealso:: :meth:`append`, :meth:`extendLeft`
        '''
        newend = self.__end + a.shape[0]
        if newend < self.__bsize:
            self.__buffer[self.__end:newend] = a
            self.__end = newend
        elif self.__bsize < self.__maxSize:
            self.resizeBuffer(min(2 * self.__bsize, self.__maxSize))
            # recursively call itself again after resizing
            return self.extend(a)
        else:
            self.moveLeft(newend - self.__bsize)
            self.__buffer[self.__end:] = a[-min(a.shape[0], self.__bsize):]
            self.__end = self.__bsize

    def extendLeft(self, a):
        ''' Prepends data to the current contents. Note that, contrary to the
        extent method, no data will be discarded if the maximum size limit is
        reached. Instead, an exception will be raised.

        :param a: (numpy.array) array of elements to append

        .. seealso:: :meth:`extend`'''
        len_a = a.shape[0]
        newend = self.__end + len_a
        if newend < self.__bsize:
            self.__buffer[len_a:newend] = self.__buffer[
                0:self.__end]  # move the contents to the right
            self.__buffer[0:len_a] = a
            self.__end = newend
        elif newend < self.__maxSize:
            self.resizeBuffer(min(2 * self.__bsize, self.__maxSize))
            # recursively call itself again after resizing
            return self.extendLeft(a)
        else:
            raise ValueError(
                'Maximum buffer size cannot be exceeded when calling extendLeft ')

    def moveLeft(self, n):
        '''discards n elements from the begginning to make space at the end of
        the buffer. Moves all elements n positions to the left and the contents
        size gets decreased by n

        **Note:** if n is larger or equal than the maximum buffer size, the
        whole buffer is wiped

        :param n: (int)'''
        newend = max(0, self.__end - n)
        self.__buffer[0:newend] = self.__buffer[n:self.__end]
        self.__end = newend

    def contents(self):
        '''returns the array of the contents that have already been filled. Note
        that it does not return the full buffer, only those elements that have
        been already set.

        It is equivalent to b[:]

        :return: (numpy.array) array of contents

        .. seealso:: :meth:`toArray`
        '''
        return self.__buffer[:self.__end]

    def toArray(self):
        '''returns a copy of the array of the contents. It is equivalent to
        ``b.contents.copy()``

        :return: (numpy.array) copy of array of contents

        .. seealso:: :meth:`contents`
        '''
        return self.contents().copy()

    def contentsSize(self):
        '''Equivalent to len(b)

        :return: (int) length of the current contents of the ArrayBuffer (not
                 the maximum size of the buffer)

        .. seealso:: :meth:`maxSize`
        '''
        return self.__end

    def bufferSize(self):
        '''Returns the current size of the internal buffer

        :return: (int) lcurrent length of the internal buffer

        .. seealso:: :meth:`contentsSize`, :meth:`maxSize`
        '''
        return self.__bsize

    def maxSize(self):
        '''Returns the maximum size of the internal buffer, beyond which the
        ArrayBuffer starts discarding elements when appending

        :return: (int) maximum length of the internal buffer

        .. seealso:: :meth:`contentsSize`, :meth:`append`, :meth:`extend`, :meth:`isFull`
        '''
        return self.__maxSize

    def setMaxSize(self, maxSize):
        '''Sets the maximum size of the internal buffer, beyond which the
        ArrayBuffer starts discarding elements when appending

        :param maxSize: (int) maximum length of the internal buffer

        .. seealso:: :meth:`contentsSize`, :meth:`append`, :meth:`extend`, :meth:`isFull`
        '''
        if maxSize < self.__bsize:
            raise ValueError(
                'Cannot set a maximum size below the current buffer size (%i)' % self.__bsize)
        self.__maxSize = maxSize

    def isFull(self):
        '''Whether the contents fill the whole of the internal buffer

        :return: (bool) True if the contents fill the maximum size. False otherwise.

        .. seealso:: :meth:`maxSize`
        '''
        return self.__end >= self.__maxSize

    def remainingSize(self):
        '''returns the remaining free space in the internal buffer (e.g., 0 if it is full)

        :return: (int) length of the unused space in the internal buffer

        .. seealso:: :meth:`contentsSize`, :meth:`maxSize`,
        '''
        return self.maxSize() - self.contentsSize()


def chunks(l, n):
    '''Generator which yields successive n-sized chunks from l'''
    for i in range(0, len(l), n):
        yield l[i:i + n]



from taurus.external.qt import Qt
from taurus.qt.qtgui.application import TaurusApplication


class MyGUI(Qt.QMainWindow):

    def __init__(self, parent=None):
        Qt.QMainWindow.__init__(self, parent)
        toolbar = self.addToolBar("Tools")

        # get icon from theme
        icon1 = Qt.QIcon.fromTheme("document-open")

        # get icon using prefix + filename
        icon2 = Qt.QIcon("actions:exit.svg")

        toolbar.addAction(icon1, "Open HDF5", self.open_file)
        toolbar.addAction(icon2, "Exit", self.close)


    def open_file(self):
        pass   # do something

if __name__ == "__main__":
    import sys
    app = TaurusApplication(cmd_line_parser=None)
    gui = MyGUI()
    gui.show()
    sys.exit(app.exec_())
